import importlib
import os
from contextlib import contextmanager
from copy import deepcopy

import lakefs_sdk
import lakefs.repository
from lakefs import config as client_config

TEST_SERVER = "https://test_server"
TEST_ACCESS_KEY_ID = "test_access_key_id"
TEST_SECRET_ACCESS_KEY = "test_secret_access_key"
TEST_CONFIG = f'''
server:
  endpoint_url: {TEST_SERVER}

credentials:
    access_key_id: {TEST_ACCESS_KEY_ID}
    secret_access_key: {TEST_SECRET_ACCESS_KEY}
'''

TEST_ENDPOINT_PATH = "/api/v1"

TEST_REPO_ARGS = lakefs_sdk.RepositoryCreation(name="test-repo",
                                               storage_namespace="s3://test_namespace",
                                               default_branch="default-branch",
                                               samples_data=True)


def get_test_client():
    from lakefs.client import Client
    clt = Client(username=TEST_ACCESS_KEY_ID, password=TEST_SECRET_ACCESS_KEY, host=TEST_SERVER)
    return clt


@contextmanager
def lakectl_test_config_context(monkey, tmp_path):
    cfg_file = tmp_path / "test.yaml"
    cfg_file.write_bytes(TEST_CONFIG.encode())
    with monkey.context():
        monkey.setattr(client_config, "_LAKECTL_YAML_PATH", cfg_file)
        from lakefs import client  # Must be imported after the monkey patching
        client = importlib.reload(client)
        try:
            yield client
        finally:
            client._DEFAULT_CLIENT = None


@contextmanager
def lakectl_no_config_context(monkey):
    with monkey.context():
        monkey.setattr(client_config, "_LAKECTL_YAML_PATH", "file_not_found")
        from lakefs import client  # Must be imported after the monkey patching
        yield client


@contextmanager
def env_var_context():
    old_env = deepcopy(os.environ)
    try:
        yield
    finally:
        os.environ = old_env


def get_test_repo() -> lakefs.Repository:
    from lakefs.client import Client
    client = Client(username="test_user", password="test_password", host="http://127.0.0.1:8000")
    return lakefs.Repository(repository_id=TEST_REPO_ARGS.name, client=client)


@contextmanager
def expect_exception_context(ex, message=None):
    try:
        yield
        assert False, f"No exception raised! Expected exception of type {ex.__name__}"
    except ex as e:
        if message is not None:
            assert message in str(e)



import pytest

try:
    from pydantic.v1 import ValidationError
except ImportError:
    from pydantic import ValidationError

import lakefs
from lakefs.exceptions import NotFoundException, TransactionException
from tests.utests.common import expect_exception_context


def test_revert(setup_repo):
    _, repo = setup_repo
    test_branch = repo.branch("main")
    initial_content = "test_content"
    test_branch.object("test_object").upload(initial_content)
    test_branch.commit("test_commit", {"test_key": "test_value"})

    override_content = "override_test_content"
    obj = test_branch.object("test_object").upload(override_content)
    test_branch.commit("override_data")

    with obj.reader(mode='r') as fd:
        assert fd.read() == override_content

    c = test_branch.revert(test_branch.head)
    assert c.message.startswith("Revert")

    with obj.reader(mode='r') as fd:
        assert fd.read() == initial_content


def test_cherry_pick(setup_repo):
    _, repo = setup_repo
    main_branch = repo.branch("main")
    test_branch = repo.branch("testest").create("main")

    initial_content = "test_content"
    test_branch.object("test_object").upload(initial_content)
    testcommit = test_branch.commit("test_commit", {"test_key": "test_value"}).get_commit()

    cherry_picked = main_branch.cherry_pick(test_branch.head)
    assert test_branch.object("test_object").exists()
    # SHAs are not equal, so we exclude them from eq checks.
    assert cherry_picked.message == testcommit.message
    # cherry-picks have origin and source ref name attached as metadata (at minimum),
    # so we only check that the additional user-supplied metadata is present.
    assert set(testcommit.metadata.items()) <= set(cherry_picked.metadata.items())
    # check that the cherry-pick origin is exactly testest@HEAD.
    assert cherry_picked.metadata["cherry-pick-origin"] == testcommit.id


def test_reset_changes(setup_repo):
    _, repo = setup_repo
    test_branch = repo.branch("main")
    paths = ["a", "b", "bar/a", "bar/b", "bar/c", "c", "foo/a", "foo/b", "foo/c", ]
    upload_data(test_branch, paths)

    validate_uncommitted_changes(test_branch, paths)

    validate_uncommitted_changes(test_branch, ["bar/a", "bar/b", "bar/c"], prefix="bar")
    test_branch.reset_changes("object", "bar/a")
    validate_uncommitted_changes(test_branch, ["a", "b", "bar/b", "bar/c", "c", "foo/a", "foo/b", "foo/c", ])

    test_branch.reset_changes("object", "bar/")

    validate_uncommitted_changes(test_branch, ["a", "b", "bar/b", "bar/c", "c", "foo/a", "foo/b", "foo/c", ])

    test_branch.reset_changes("common_prefix", "foo/")
    validate_uncommitted_changes(test_branch, ["a", "b", "bar/b", "bar/c", "c"])

    test_branch.reset_changes()
    validate_uncommitted_changes(test_branch, [])


def test_delete_object_changes(setup_repo):
    _, repo = setup_repo
    test_branch = repo.branch("main")
    path_and_data = ["a", "b", "bar/a", "bar/b", "bar/c", "c", "foo/a", "foo/b", "foo/c"]
    upload_data(test_branch, path_and_data)
    test_branch.commit("add some files", {"test_key": "test_value"})

    test_branch.delete_objects("foo/a")
    validate_uncommitted_changes(test_branch, ["foo/a"], "removed")

    paths = {"foo/b", "foo/c"}
    test_branch.delete_objects(paths)
    validate_uncommitted_changes(test_branch, ["foo/a", "foo/b", "foo/c"], "removed")
    repo = lakefs.Repository(test_branch.repo_id)
    test_branch.delete_objects([repo.ref(test_branch.head).object("a"), repo.ref(test_branch.head).object("b")])
    validate_uncommitted_changes(test_branch, ["a", "b", "foo/a", "foo/b", "foo/c"], "removed")
    with expect_exception_context(ValidationError):
        test_branch.reset_changes("unknown", "foo/")


def upload_data(branch, path_and_data, multiplier=1):
    for s in path_and_data:
        branch.object(s).upload(s * multiplier)


def validate_uncommitted_changes(branch, expected, change_type="added", prefix=""):
    count = 0
    for index, change in enumerate(branch.uncommitted(max_amount=10, prefix=prefix)):
        assert change.path == expected[index]
        assert change.path_type == "object"
        assert change.type == change_type
        assert change.size_bytes == 0 if change_type == "removed" else len(expected[index])
        count += 1
    if count != len(expected):
        raise AssertionError(f"Expected {len(expected)} changes, got {count}")


def test_transaction(setup_repo):
    _, repo = setup_repo
    path_and_data1 = ["a", "b", "bar/a", "bar/b", "bar/c", "c"]
    path_and_data2 = ["foo/a", "foo/b", "foo/c"]
    test_branch = repo.branch("main")

    with test_branch.transact(commit_message="my transaction", commit_metadata={"foo": "bar"}) as tx:
        assert tx.get_commit().id == test_branch.head.id
        upload_data(tx, path_and_data1)
        upload_data(tx, path_and_data2)
        tx.reset_changes(path_type="common_prefix", path="foo")
        tx_id = tx.id

    # Verify transaction branch was deleted
    with expect_exception_context(NotFoundException):
        repo.branch(tx.id).get_commit()

    #  Verify transaction completed successfully
    log = list(test_branch.log(amount=2))
    assert log[0].message == f"Merge transaction {tx_id} to branch"
    assert log[1].message == "my transaction"
    assert log[1].metadata.get("foo") == "bar"

    for obj in path_and_data1:
        assert test_branch.object(obj).exists()

    for obj in path_and_data2:
        assert not test_branch.object(obj).exists()

    # Reset all changes - ensure no new commits
    with expect_exception_context(TransactionException, "no changes"):
        with test_branch.transact(commit_message="my transaction", commit_metadata={"foo": "bar"}) as tx:
            assert tx.get_commit().id == test_branch.head.id
            upload_data(tx, path_and_data2)
            tx.reset_changes()

    log = list(test_branch.log(amount=1))
    assert log[0].message == f"Merge transaction {tx_id} to branch"

    # Verify transaction branch is deleted when no changes are made
    with expect_exception_context(TransactionException, "no changes"):
        with test_branch.transact(commit_message="my transaction") as tx:
            pass

    with expect_exception_context(NotFoundException):
        tx.get_commit()


@pytest.mark.parametrize("cleanup_branch", [True, False])
def test_transaction_failure(setup_repo, cleanup_branch):
    _, repo = setup_repo
    new_data = ["a", "b", "bar/a", "bar/b"]
    common_data = ["foo/a", "foo/b", "foo/c"]
    test_branch = repo.branch("main")
    commit = test_branch.get_commit()

    # Exception during transaction
    with expect_exception_context(ValueError, "Something bad happened"):
        with test_branch.transact(commit_message="my transaction", delete_branch_on_error=cleanup_branch) as tx:
            upload_data(tx, new_data)
            raise ValueError("Something bad happened")

    # Ensure tx branch exists and not merged
    if not cleanup_branch:
        assert tx.get_commit()

    assert test_branch.get_commit() == commit

    # Merge on dirty branch
    with expect_exception_context(TransactionException, "dirty branch"):
        with test_branch.transact(commit_message="my transaction") as tx:
            assert tx.get_commit().id == test_branch.head.id
            upload_data(tx, new_data)
            upload_data(test_branch, common_data, 2)

    # Ensure tx branch exists and not merged
    if not cleanup_branch:
        assert tx.get_commit()

    assert test_branch.get_commit() == commit

    # Merge with conflicts
    with expect_exception_context(TransactionException, "Conflict"):
        with test_branch.transact(commit_message="my transaction") as tx:
            new_ref = test_branch.commit(message="test branch commit")
            upload_data(tx, common_data)

    with expect_exception_context(NotFoundException):
        assert tx.get_commit()

    assert test_branch.get_commit() == new_ref.get_commit()



from lakefs.exceptions import NoAuthenticationFound
from tests.utests.common import (
    lakectl_test_config_context,
    lakectl_no_config_context,
    TEST_SERVER,
    TEST_ACCESS_KEY_ID,
    TEST_SECRET_ACCESS_KEY,
    TEST_ENDPOINT_PATH, expect_exception_context
)

TEST_CONFIG_KWARGS: dict[str, str] = {
    "username": "my_username",
    "password": "my_password",
    "host": "http://my_host",
    "access_token": "my_jwt_token"
}


class TestClient:
    def test_client_no_config(self, monkeypatch):
        with lakectl_no_config_context(monkeypatch) as client:
            with expect_exception_context(NoAuthenticationFound):
                client.Client()

    def test_client_no_kwargs(self, monkeypatch, tmp_path):
        with lakectl_test_config_context(monkeypatch, tmp_path) as client:
            clt = client.Client()
            config = clt.config
            assert config.host == TEST_SERVER + TEST_ENDPOINT_PATH
            assert config.username == TEST_ACCESS_KEY_ID
            assert config.password == TEST_SECRET_ACCESS_KEY

    def test_client_kwargs(self, monkeypatch, tmp_path):
        # Use lakectl yaml file and ensure it is not being read in case kwargs are provided
        with lakectl_test_config_context(monkeypatch, tmp_path) as client:
            clt = client.Client(**TEST_CONFIG_KWARGS)
            config = clt.config
            assert config.host == TEST_CONFIG_KWARGS["host"] + TEST_ENDPOINT_PATH
            assert config.username == TEST_CONFIG_KWARGS["username"]
            assert config.password == TEST_CONFIG_KWARGS["password"]
            assert config.access_token == TEST_CONFIG_KWARGS["access_token"]



"""
Exceptions module
"""
import http
import json
from contextlib import contextmanager
from typing import Optional, Callable

import lakefs_sdk.exceptions
from urllib3 import HTTPResponse


class LakeFSException(Exception):
    """
    Base exception for all SDK exceptions
    """


class ServerException(LakeFSException):
    """
    Generic exception when no other exception is applicable
    """
    status_code: int
    reason: str
    body: dict

    def __init__(self, status=None, reason=None, body=None):
        self.status_code = status
        self.reason = reason
        if body is not None:
            try:  # Try to get message from body
                self.body = json.loads(body)
            except json.JSONDecodeError:
                self.body = {}
        else:
            self.body = {}

    def __str__(self):
        return f"code: {self.status_code}, reason: {self.reason}, body: {self.body}"


class NotFoundException(ServerException):
    """
    Resource could not be found on lakeFS server
    """


class ForbiddenException(ServerException):
    """
    Operation not permitted
    """


class NoAuthenticationFound(LakeFSException):
    """
    Raised when no authentication method could be found on Client instantiation
    """


class BadRequestException(ServerException):
    """
    Bad Request
    """


class NotAuthorizedException(ServerException):
    """
    User not authorized to perform operation
    """


class UnsupportedOperationException(ServerException):
    """
    Operation not supported by lakeFS server or SDK
    """


class ConflictException(ServerException):
    """
     Resource / request conflict
    """


class ObjectNotFoundException(NotFoundException, FileNotFoundError):
    """
    Raised when the currently used object no longer exist in the lakeFS server
    """


class ObjectExistsException(ServerException, FileExistsError):
    """
    Raised when Object('...').create(mode='x') and object exists
    """


class PermissionException(NotAuthorizedException, PermissionError):
    """
    Raised by Object.open() and Object.create() for compatibility with python
    """


class InvalidRangeException(ServerException, OSError):
    """
    Raised when the reference could not be found in the lakeFS server
    """


class ImportManagerException(LakeFSException):
    """
    Import manager exceptions that are not originated from the SDK
    """


class TransactionException(LakeFSException):
    """
    Exceptions during the transaction commit logic
    """


_STATUS_CODE_TO_EXCEPTION = {
    http.HTTPStatus.BAD_REQUEST.value: BadRequestException,
    http.HTTPStatus.UNAUTHORIZED.value: NotAuthorizedException,
    http.HTTPStatus.FORBIDDEN.value: ForbiddenException,
    http.HTTPStatus.NOT_FOUND.value: NotFoundException,
    http.HTTPStatus.METHOD_NOT_ALLOWED.value: UnsupportedOperationException,
    http.HTTPStatus.CONFLICT.value: ConflictException,
    http.HTTPStatus.REQUESTED_RANGE_NOT_SATISFIABLE.value: InvalidRangeException
}


@contextmanager
def api_exception_handler(custom_handler: Optional[Callable[[LakeFSException], LakeFSException]] = None):
    """
    Contexts which converts lakefs_sdk API exceptions to LakeFS exceptions and handles them.

    :param custom_handler: Optional handler which can be used to provide custom behavior for specific exceptions.
        If custom_handler returns an exception, this function will raise the exception at the end of the
        custom_handler invocation.
    """
    try:
        yield
    except lakefs_sdk.ApiException as e:
        lakefs_ex = _STATUS_CODE_TO_EXCEPTION.get(e.status, ServerException)(e.status, e.reason, e.body)
        if custom_handler is not None:
            lakefs_ex = custom_handler(lakefs_ex)

        if lakefs_ex is not None:
            raise lakefs_ex from e


def handle_http_error(resp: HTTPResponse) -> None:
    """
    Handles http response and raises the appropriate lakeFS exception if needed

    :param resp: The response to parse
    """
    if not http.HTTPStatus.OK <= resp.status < http.HTTPStatus.MULTIPLE_CHOICES:
        lakefs_ex = _STATUS_CODE_TO_EXCEPTION.get(resp.status, ServerException)(resp.status, resp.reason, resp.data)
        raise lakefs_ex



import http
from contextlib import contextmanager
from typing import get_args
import urllib3

import pytest

import lakefs_sdk.api

from lakefs.object import ReadModes
from tests.utests.common import get_test_client, expect_exception_context


class ObjectTestKWArgs:
    def __init__(self) -> None:
        self.repository_id = "test_repo"
        self.reference_id = "test_reference"
        self.path = "test_path"


class StorageTestConfig(lakefs_sdk.StorageConfig):

    def __init__(self) -> None:
        super().__init__(blockstore_type="s3",
                         blockstore_namespace_example="",
                         blockstore_namespace_ValidityRegex="",
                         pre_sign_support=True,
                         pre_sign_support_ui=False,
                         import_support=False,
                         import_validity_regex="")


class ObjectTestStats(lakefs_sdk.ObjectStats):
    def __init__(self) -> None:
        super().__init__(path="",
                         path_type="object",
                         physical_address="",
                         checksum="",
                         mtime=0)


class StagingTestLocation(lakefs_sdk.StagingLocation):
    def __init__(self) -> None:
        super().__init__(physical_address="physical_address")


@contextmanager
def readable_object_context(monkey, **kwargs):
    with monkey.context():
        from lakefs.object import StoredObject
        clt = get_test_client()
        conf = lakefs_sdk.Config(version_config=lakefs_sdk.VersionConfig(), storage_config=StorageTestConfig())
        monkey.setattr(clt, "_server_conf", conf)
        read_obj = StoredObject(client=clt, **kwargs)
        yield read_obj


@contextmanager
def writeable_object_context(monkey, **kwargs):
    with monkey.context():
        monkey.setattr(lakefs_sdk.api.BranchesApi, "get_branch", lambda *args: None)
        from lakefs.object import WriteableObject
        conf = lakefs_sdk.Config(version_config=lakefs_sdk.VersionConfig(), storage_config=StorageTestConfig())
        clt = get_test_client()
        monkey.setattr(clt, "_server_conf", conf)
        obj = WriteableObject(client=clt, **kwargs)
        yield obj


class TestStoredObject:
    def test_exists(self, monkeypatch, tmp_path):
        test_kwargs = ObjectTestKWArgs()
        with readable_object_context(monkeypatch, **test_kwargs.__dict__) as obj:
            # Object exists
            monkeypatch.setattr(lakefs_sdk.api.ObjectsApi, "head_object", lambda *args: None)
            assert obj.exists()
            # Object doesn't exist
            monkeypatch.setattr(lakefs_sdk.api.ObjectsApi, "head_object",
                                lambda *args: (_ for _ in ()).throw(lakefs_sdk.exceptions.NotFoundException(
                                    status=http.HTTPStatus.NOT_FOUND)))
            assert not obj.exists()

            # Other exception
            monkeypatch.setattr(lakefs_sdk.api.ObjectsApi, "head_object",
                                lambda *args: 1 / 0)
            with expect_exception_context(ZeroDivisionError):
                obj.exists()


class TestObjectReader:
    def test_seek(self, monkeypatch, tmp_path):
        test_kwargs = ObjectTestKWArgs()
        with readable_object_context(monkeypatch, **test_kwargs.__dict__) as obj:
            with obj.reader() as fd:
                assert fd.tell() == 0
                fd.seek(30)
                assert fd.tell() == 30
                with expect_exception_context(OSError):
                    fd.seek(-1)

            # Create another reader
            with obj.reader() as fd:
                assert fd.tell() == 0

    def test_fileno(self, monkeypatch, tmp_path):
        test_kwargs = ObjectTestKWArgs()
        with readable_object_context(monkeypatch, **test_kwargs.__dict__) as obj:
            with obj.reader() as fd:
                with expect_exception_context(OSError):
                    fd.fileno()

    @staticmethod
    def verify_reader(fd, patch_setattr, test_kwargs, data):
        object_stats = ObjectTestStats()
        object_stats.path = test_kwargs.path
        object_stats.size_bytes = len(data)
        patch_setattr(lakefs_sdk.api.ObjectsApi, "stat_object", lambda *args: object_stats)

        # read negative
        with expect_exception_context(OSError):
            fd.read(-1)

        # Read whole file
        start_pos = 0
        end_pos = ""

        def monkey_get_object(_, repository, ref, path, range, presign, **__):  # pylint: disable=W0622
            assert repository == test_kwargs.repository_id
            assert ref == test_kwargs.reference_id
            assert path == test_kwargs.path
            assert presign

            if isinstance(end_pos, int):
                return data[start_pos:end_pos]
            return data[start_pos:]

        patch_setattr(lakefs_sdk.api.ObjectsApi, "get_object", monkey_get_object)
        assert fd.read() == data
        assert fd.tell() == object_stats.size_bytes

        # Test reading from middle
        start_pos = 132
        fd.seek(start_pos)
        read_size = 456
        end_pos = start_pos + read_size - 1
        fd.read(read_size)
        assert fd.tell() == start_pos + read_size - 1

        # Read more than file size
        start_pos = fd.tell()
        read_size = 2 * object_stats.size_bytes
        end_pos = start_pos + 2 * object_stats.size_bytes - 1
        fd.read(read_size)
        assert fd.tell() == object_stats.size_bytes

    def test_read_by_context(self, monkeypatch, tmp_path):
        test_kwargs = ObjectTestKWArgs()
        with readable_object_context(monkeypatch, **test_kwargs.__dict__) as obj:
            data = b"test \xcf\x84o\xcf\x81\xce\xbdo\xcf\x82\n" * 100
            with obj.reader(mode="rb") as fd:
                self.verify_reader(fd, monkeypatch.setattr, test_kwargs, data)

    def test_read_by_calling_reader(self, monkeypatch, tmp_path):
        test_kwargs = ObjectTestKWArgs()
        with readable_object_context(monkeypatch, **test_kwargs.__dict__) as obj:
            data = b"test \xcf\x84o\xcf\x81\xce\xbdo\xcf\x82\n" * 100
            fd = obj.reader(mode="rb")
            self.verify_reader(fd, monkeypatch.setattr, test_kwargs, data)
            fd.close()
            assert fd.closed

    @pytest.mark.parametrize("mode", [*get_args(ReadModes)])
    def test_read_modes(self, monkeypatch, tmp_path, mode):
        test_kwargs = ObjectTestKWArgs()
        data = b"test \xcf\x84o\xcf\x81\xce\xbdo\xcf\x82"
        with readable_object_context(monkeypatch, **test_kwargs.__dict__) as obj:
            with obj.reader(mode=mode) as fd:
                object_stats = ObjectTestStats()
                object_stats.path = test_kwargs.path
                object_stats.size_bytes = len(data)
                monkeypatch.setattr(lakefs_sdk.api.ObjectsApi, "stat_object", lambda *args: object_stats)

                # Read whole file
                start_pos = 0

                def monkey_get_object(_, repository, ref, path, range, presign, **__):  # pylint: disable=W0622
                    assert repository == test_kwargs.repository_id
                    assert ref == test_kwargs.reference_id
                    assert path == test_kwargs.path
                    assert range is None
                    assert presign
                    return b"test \xcf\x84o\xcf\x81\xce\xbdo\xcf\x82"

                monkeypatch.setattr(lakefs_sdk.api.ObjectsApi, "get_object", monkey_get_object)
                res = fd.read()
                if 'b' not in mode:
                    assert res == data.decode('utf-8')
                else:
                    assert res == data

                assert fd.tell() == start_pos + object_stats.size_bytes

    def test_read_invalid_mode(self, monkeypatch, tmp_path):
        test_kwargs = ObjectTestKWArgs()
        with readable_object_context(monkeypatch, **test_kwargs.__dict__) as obj:
            with expect_exception_context(ValueError):
                with obj.reader(mode="invalid"):
                    pass


class TestWriteableObject:
    def test_upload(self, monkeypatch, tmp_path):
        test_kwargs = ObjectTestKWArgs()
        with writeable_object_context(monkeypatch, **test_kwargs.__dict__) as obj:
            staging_location = StagingTestLocation()
            monkeypatch.setattr(lakefs_sdk.api.StagingApi, "get_physical_address", lambda *args: staging_location)
            monkeypatch.setattr(urllib3.PoolManager, "request",
                                lambda *args, **kwargs: urllib3.response.HTTPResponse(status=201))

            def monkey_link_physical_address(*_, staging_metadata: lakefs_sdk.StagingMetadata, **__):
                assert staging_metadata.size_bytes == len(data)
                assert staging_metadata.staging == staging_location
                return lakefs_sdk.ObjectStats(path=obj.path,
                                              path_type="object",
                                              physical_address=staging_location.physical_address,
                                              checksum="",
                                              mtime=12345)

            monkeypatch.setattr(lakefs_sdk.api.StagingApi, "link_physical_address", monkey_link_physical_address)
            # Test string
            data = "test_data"
            obj.upload(data=data)

    def test_upload_invalid_mode(self, monkeypatch, tmp_path):
        test_kwargs = ObjectTestKWArgs()
        with writeable_object_context(monkeypatch, **test_kwargs.__dict__) as obj:
            with expect_exception_context(ValueError):
                obj.upload(data="", mode="invalid")


class TestObjectWriter:
    def test_fileno(self, monkeypatch, tmp_path):
        test_kwargs = ObjectTestKWArgs()
        with writeable_object_context(monkeypatch, **test_kwargs.__dict__) as obj:
            with obj.reader() as fd:
                with expect_exception_context(OSError):
                    fd.fileno()



"""
Module containing lakeFS reference implementation
"""

from __future__ import annotations

from typing import Optional, Generator, Union

import lakefs_sdk

from lakefs.models import Commit, Change, CommonPrefix, ObjectInfo, _OBJECT
from lakefs.client import Client, _BaseLakeFSObject
from lakefs.exceptions import api_exception_handler
from lakefs.object import StoredObject


class Reference(_BaseLakeFSObject):
    """
    Class representing a reference in lakeFS.
    """
    _repo_id: str
    _id: str
    _commit: Optional[Commit] = None

    def __init__(self, repository_id: str, reference_id: str, client: Optional[Client] = None) -> None:
        """Return a reference to a lakeFS commit.

        :param repository_id: the repository holding the commit
        :param reference_id: a reference expression to the commit

        Any reference expression can be used as a reference_id, for example:

        - 'main' (head of 'main' branch)
        - 'main@' (head of 'main' branch, only committed objects)
        - 'my_tag~3' (3 commits before 'my_tag')

        See https://docs.lakefs.io/understand/model.html#ref-expressions for
        details.
        """
        self._repo_id = repository_id
        self._id = reference_id
        super().__init__(client)

    @property
    def repo_id(self) -> str:
        """
        Return the repository id for this reference
        """
        return self._repo_id

    @property
    def id(self) -> str:
        """
        Returns the reference id
        """
        return self._id

    def objects(self,
                max_amount: Optional[int] = None,
                after: Optional[str] = None,
                prefix: Optional[str] = None,
                delimiter: Optional[str] = None,
                **kwargs) -> Generator[StoredObject | CommonPrefix]:
        """
        Returns an object generator for this reference, the generator can yield either a StoredObject or a CommonPrefix
        object depending on the listing parameters provided.

        :param max_amount: Stop showing changes after this amount
        :param after: Return items after this value
        :param prefix: Return items prefixed with this value
        :param delimiter: Group common prefixes by this delimiter
        :param kwargs: Additional Keyword Arguments to send to the server
        :raise NotFoundException: if this reference or other_ref does not exist
        :raise NotAuthorizedException: if user is not authorized to perform this operation
        :raise ServerException: for any other errors
        """

        for res in generate_listing(self._client.sdk_client.objects_api.list_objects,
                                    repository=self._repo_id,
                                    ref=self._id,
                                    max_amount=max_amount,
                                    after=after,
                                    prefix=prefix,
                                    delimiter=delimiter,
                                    **kwargs):
            type_class = ObjectInfo if res.path_type == _OBJECT else CommonPrefix
            yield type_class(**res.dict())

    def log(self, max_amount: Optional[int] = None, **kwargs) -> Generator[Commit]:
        """
        Returns a generator of commits starting with this reference id

        :param max_amount: (Optional) limits the amount of results to return from the server
        :param kwargs: Additional Keyword Arguments to send to the server
        :raise NotFoundException: if reference by this id does not exist
        :raise NotAuthorizedException: if user is not authorized to perform this operation
        :raise ServerException: for any other errors
        """
        for res in generate_listing(self._client.sdk_client.refs_api.log_commits, self._repo_id, self._id,
                                    max_amount=max_amount, **kwargs):
            yield Commit(**res.dict())

    def get_commit(self) -> Commit:
        """
        Returns the underlying commit referenced by this reference id

        :raise NotFoundException: if this reference does not exist
        :raise NotAuthorizedException: if user is not authorized to perform this operation
        :raise ServerException: for any other errors
        """
        if self._commit is None:
            with api_exception_handler():
                commit = self._client.sdk_client.commits_api.get_commit(self._repo_id, self._id)
                self._commit = Commit(**commit.dict())
        return self._commit

    def diff(self,
             other_ref: ReferenceType,
             max_amount: Optional[int] = None,
             after: Optional[str] = None,
             prefix: Optional[str] = None,
             delimiter: Optional[str] = None,
             **kwargs) -> Generator[Change]:
        """
        Returns a diff generator of changes between this reference and other_ref

        :param other_ref: The other ref to diff against
        :param max_amount: Stop showing changes after this amount
        :param after: Return items after this value
        :param prefix: Return items prefixed with this value
        :param delimiter: Group common prefixes by this delimiter
        :param kwargs: Additional Keyword Arguments to send to the server
        :raise NotFoundException: if this reference or other_ref does not exist
        :raise NotAuthorizedException: if user is not authorized to perform this operation
        :raise ServerException: for any other errors
        """
        other_ref_id = other_ref if isinstance(other_ref, str) else other_ref.id
        for diff in generate_listing(self._client.sdk_client.refs_api.diff_refs,
                                     repository=self._repo_id,
                                     left_ref=self._id,
                                     right_ref=other_ref_id,
                                     after=after,
                                     max_amount=max_amount,
                                     prefix=prefix,
                                     delimiter=delimiter,
                                     **kwargs):
            yield Change(**diff.dict())

    def merge_into(self, destination_branch: ReferenceType, **kwargs) -> str:
        """
        Merge this reference into destination branch

        :param destination_branch: The merge destination (either ID or branch object)
        :param kwargs: Additional Keyword Arguments to send to the server
        :return: The reference id of the merge commit
        :raise NotFoundException: if reference by this id does not exist, or branch doesn't exist
        :raise NotAuthorizedException: if user is not authorized to perform this operation
        :raise ServerException: for any other errors
        """
        branch_id = destination_branch if isinstance(destination_branch, str) else destination_branch.id
        with api_exception_handler():
            merge = lakefs_sdk.Merge(**kwargs)
            res = self._client.sdk_client.refs_api.merge_into_branch(self._repo_id,
                                                                     self._id,
                                                                     branch_id,
                                                                     merge=merge)
            return res.reference

    def object(self, path: str) -> StoredObject:  # pylint: disable=C0103
        """
        Returns an Object class representing a lakeFS object with this repo id, reference id and path

        :param path: The object's path
        """
        return StoredObject(self._repo_id, self._id, path, self._client)

    def __repr__(self):
        class_name = self.__class__.__name__
        return f'{class_name}(repository="{self.repo_id}", id="{self.id}")'


def generate_listing(func, *args, max_amount: Optional[int] = None, **kwargs):
    """
    Generic generator function, for lakefs-sdk listings functionality

    :param func: The listing function
    :param args: The function args
    :param max_amount: The max amount of objects to generate
    :param kwargs: The function kwargs
    :return: A generator based on the listing function
    """
    has_more = True
    with api_exception_handler():
        while has_more:
            page = func(*args, **kwargs)
            has_more = page.pagination.has_more
            kwargs["after"] = page.pagination.next_offset
            for res in page.results:
                yield res

                if max_amount is not None:
                    max_amount -= 1
                    if max_amount <= 0:
                        return


ReferenceType = Union[str, Reference, Commit]



"""
Module containing lakeFS repository implementation
"""

from __future__ import annotations

from typing import Optional, Generator

import lakefs_sdk

from lakefs.models import RepositoryProperties
from lakefs.tag import Tag
from lakefs.branch import Branch
from lakefs.client import Client, _BaseLakeFSObject
from lakefs.exceptions import api_exception_handler, ConflictException, LakeFSException
from lakefs.reference import Reference, generate_listing


class Repository(_BaseLakeFSObject):
    """
    Class representing a Repository in lakeFS.
    The Repository object provides the context for the other objects that are found in it.
    Access to these objects should be done from this class
    """
    _id: str
    _properties: RepositoryProperties = None

    def __init__(self, repository_id: str, client: Optional[Client] = None) -> None:
        self._id = repository_id
        super().__init__(client)

    def create(self,
               storage_namespace: str,
               default_branch: str = "main",
               include_samples: bool = False,
               exist_ok: bool = False,
               **kwargs) -> Repository:
        """
        Create a new repository in lakeFS from this object

        :param storage_namespace: Repository's storage namespace
        :param default_branch: The default branch for the repository. If None, use server default name
        :param include_samples: Whether to include sample data in repository creation
        :param exist_ok: If False will throw an exception if a repository by this name already exists. Otherwise,
            return the existing repository without creating a new one
        :param kwargs: Additional Keyword Arguments to send to the server
        :return: The lakeFS SDK object representing the repository
        :raise NotAuthorizedException: if user is not authorized to perform this operation
        :raise ServerException: for any other errors
        """
        repository_creation = lakefs_sdk.RepositoryCreation(name=self._id,
                                                            storage_namespace=storage_namespace,
                                                            default_branch=default_branch,
                                                            sample_data=include_samples)

        def handle_conflict(e: LakeFSException):
            if isinstance(e, ConflictException) and exist_ok:
                with api_exception_handler():
                    get_repo = self._client.sdk_client.repositories_api.get_repository(self._id)
                    self._properties = RepositoryProperties(**get_repo.dict())
                    return None
            return e

        with api_exception_handler(handle_conflict):
            repo = self._client.sdk_client.repositories_api.create_repository(repository_creation, **kwargs)
            self._properties = RepositoryProperties(**repo.dict())
        return self

    def delete(self) -> None:
        """
        Delete repository from lakeFS server

        :raise NotFoundException: if repository by this id does not exist
        :raise NotAuthorizedException: if user is not authorized to perform this operation
        :raise ServerException: for any other errors
        """
        with api_exception_handler():
            self._client.sdk_client.repositories_api.delete_repository(self._id)

    def branch(self, branch_id: str) -> Branch:
        """
        Return a branch object using the current repository id and client
        :param branch_id: name of the branch
        """
        return Branch(self._id, branch_id, self._client)

    def commit(self, commit_id: str) -> Reference:
        """
        Return a reference object using the current repository id and client
        :param commit_id: id of the commit reference
        """
        return Reference(self._id, commit_id, self._client)

    def ref(self, ref_id: str) -> Reference:
        """
        Return a reference object using the current repository id and client
        :param ref_id: branch name, commit id or tag id
        """
        return Reference(self._id, ref_id, self._client)

    def tag(self, tag_id: str) -> Tag:
        """
        Return a tag object using the current repository id and client
        :param tag_id: name of the tag
        """
        return Tag(self._id, tag_id, self._client)

    def branches(self, max_amount: Optional[int] = None,
                 after: Optional[str] = None, prefix: Optional[str] = None, **kwargs) -> Generator[Branch]:
        """
        Returns a generator listing for branches on the given repository

        :param max_amount: Stop showing changes after this amount
        :param after: Return items after this value
        :param prefix: Return items prefixed with this value
        :param kwargs: Additional Keyword Arguments to send to the server
        :raise NotFoundException: if repository does not exist
        :raise NotAuthorizedException: if user is not authorized to perform this operation
        :raise ServerException: for any other errors
        """

        for res in generate_listing(self._client.sdk_client.branches_api.list_branches, self._id,
                                    max_amount=max_amount, after=after, prefix=prefix, **kwargs):
            yield Branch(self._id, res.id, client=self._client)

    def tags(self, max_amount: Optional[int] = None,
             after: Optional[str] = None, prefix: Optional[str] = None, **kwargs) -> Generator[Tag]:
        """
        Returns a generator listing for tags on the given repository

        :param max_amount: Stop showing changes after this amount
        :param after: Return items after this value
        :param prefix: Return items prefixed with this value
        :param kwargs: Additional Keyword Arguments to send to the server
        :raise NotFoundException: if repository does not exist
        :raise NotAuthorizedException: if user is not authorized to perform this operation
        :raise ServerException: for any other errors
        """
        for res in generate_listing(self._client.sdk_client.tags_api.list_tags, self._id,
                                    max_amount=max_amount, after=after, prefix=prefix, **kwargs):
            yield Tag(self._id, res.id, client=self._client)

    @property
    def metadata(self) -> dict[str, str]:
        """
        Returns the repository metadata
        """
        with api_exception_handler():
            return self._client.sdk_client.repositories_api.get_repository_metadata(repository=self._id)

    @property
    def properties(self) -> RepositoryProperties:
        """
        Return the repository's properties object
        """
        if self._properties is None:
            with api_exception_handler():
                repo = self._client.sdk_client.repositories_api.get_repository(self._id)
                self._properties = RepositoryProperties(**repo.dict())
                return self._properties

        return self._properties

    @property
    def id(self) -> str:
        """
        Returns the repository's id
        """
        return self._id

    def __repr__(self) -> str:
        return f'Repository(id="{self.id}")'

    def __str__(self):
        return str(self.properties)


def repositories(client: Client = None,
                 prefix: Optional[str] = None,
                 after: Optional[str] = None,
                 **kwargs) -> Generator[Repository]:
    """
    Creates a repositories object generator listing lakeFS repositories

    :param client: The lakeFS client to use, if None, tries to use the default client
    :param prefix: Return items prefixed with this value
    :param after: Return items after this value
    :return: A generator listing lakeFS repositories
    """
    if client is None:  # Try to get default client
        client = Client()

    for res in generate_listing(client.sdk_client.repositories_api.list_repositories,
                                prefix=prefix,
                                after=after,
                                **kwargs):
        yield Repository(res.id, client)



"""
Module containing lakeFS tag implementation
"""

from __future__ import annotations
from typing import Optional

import lakefs_sdk

from lakefs.client import Client
from lakefs.exceptions import api_exception_handler, LakeFSException, ConflictException
from lakefs.reference import Reference, ReferenceType


class Tag(Reference):
    """
    Class representing a tag in lakeFS.
    """

    def __init__(self, repository_id: str, tag_id: str, client: Optional[Client] = None):
        super().__init__(repository_id, reference_id=tag_id, client=client)

    def create(self, source_ref: ReferenceType, exist_ok: Optional[bool] = False) -> Tag:
        """
        Create a tag from the given source_ref

        :param source_ref: The reference to create the tag on (either ID or Reference object)
        :param exist_ok: If True returns the existing Tag reference otherwise raises exception
        :return: A lakeFS SDK Tag object
        :raise NotAuthorizedException: if user is not authorized to perform this operation
        :raise NotFoundException: if source_ref_id doesn't exist on the lakeFS server
        :raise ServerException: for any other errors.
        """
        source_ref_id = source_ref if isinstance(source_ref, str) else source_ref.id
        tag_creation = lakefs_sdk.TagCreation(id=self.id, ref=source_ref_id)

        def handle_conflict(e: LakeFSException):
            if not (isinstance(e, ConflictException) and exist_ok):
                return e
            return None

        with api_exception_handler(handle_conflict):
            self._client.sdk_client.tags_api.create_tag(self._repo_id, tag_creation)

        return self

    def delete(self) -> None:
        """
        Delete the tag from the lakeFS server

        :raise NotAuthorizedException: if user is not authorized to perform this operation
        :raise NotFoundException: if source_ref_id doesn't exist on the lakeFS server
        :raise ServerException: for any other errors
        """
        with api_exception_handler():
            self._client.sdk_client.tags_api.delete_tag(self._repo_id, self.id)
            self._commit = None



import io
import csv
import json
import math
import os
from xml.etree import ElementTree
from typing import get_args

import pandas as pd
import yaml
import pytest
from PIL import Image

from tests.integration.conftest import TEST_DATA
from tests.utests.common import expect_exception_context
from lakefs.exceptions import ObjectExistsException, NotFoundException
from lakefs.object import WriteableObject, WriteModes, ReadModes


@pytest.mark.parametrize("pre_sign", (True, False), indirect=True)
def test_object_read_seek(setup_repo, pre_sign):
    clt, repo = setup_repo
    data = b"test_data"
    obj = WriteableObject(repository_id=repo.properties.id, reference_id="main", path="test_obj", client=clt).upload(
        data=data, pre_sign=pre_sign)

    with obj.reader() as fd:
        assert fd.read(2 * len(data)) == data
        fd.seek(2)

        assert fd.read(5) == data[2:7]

        assert fd.read() == data[7:]

        # This should return an empty string (simulates behavior of builtin open())
        assert len(fd.read(1)) == 0

        fd.seek(0)
        for c in data:
            assert ord(fd.read(1)) == c
        # This should return an empty string (simulates behavior of builtin open())
        assert len(fd.read(1)) == 0

        fd.seek(-4, os.SEEK_END)
        assert fd.read() == data[len(data) - 4:]

        with expect_exception_context(OSError):
            fd.seek(-len(data) - 1, os.SEEK_CUR)

        assert fd.tell() == len(data)

        fd.seek(-4, os.SEEK_CUR)
        assert fd.read() == data[len(data) - 4:]

        with expect_exception_context(io.UnsupportedOperation):
            fd.seek(0, 10)

    assert fd.closed

    # read after close
    with expect_exception_context(ValueError, "I/O operation on closed file"):
        fd.read(10)

    # seek after close
    with expect_exception_context(ValueError, "I/O operation on closed file"):
        fd.seek(10)

    # readline after close
    with expect_exception_context(ValueError, "I/O operation on closed file"):
        fd.readline()

    # flush after close
    with expect_exception_context(ValueError, "I/O operation on closed file"):
        fd.flush()

    # Close a second time should not fail
    fd.close()


def test_object_upload_exists(setup_repo):
    clt, repo = setup_repo
    data = b"test_data"
    obj = WriteableObject(repository_id=repo.properties.id, reference_id="main", path="test_obj", client=clt).upload(
        data=data)
    with expect_exception_context(ObjectExistsException):
        obj.upload(data="some_other_data", mode='xb')

    with obj.reader() as fd:
        assert fd.read() == data

    # Create - overwrite
    new_data = b"new_data"
    obj2 = obj.upload(data=new_data, mode='w')

    with obj.reader() as fd:
        assert fd.read() == new_data

    assert obj2 == obj


@pytest.mark.parametrize("w_mode", get_args(WriteModes))
@pytest.mark.parametrize("r_mode", get_args(ReadModes))
@pytest.mark.parametrize("pre_sign", (True, False), indirect=True)
def test_object_upload_read_different_params(setup_repo, w_mode, r_mode, pre_sign):
    clt, repo = setup_repo

    # urllib3 encodes TextIOBase to ISO-8859-1, data should contain symbols that can be encoded as such
    data = b'test \x48\x65\x6c\x6c\x6f\x20\x57\x6f\x72\x6c\x64\x21'
    obj = WriteableObject(repository_id=repo.properties.id, reference_id="main", path="test_obj", client=clt).upload(
        data=data, mode=w_mode, pre_sign=pre_sign)

    with obj.reader(mode=r_mode) as fd:
        res = fd.read()
        if 'b' in r_mode:
            assert res == data
        else:
            assert res.encode('utf-8') == data


def test_object_copy(setup_repo):
    clt, repo = setup_repo
    data = "test_data"
    obj = WriteableObject(repository_id=repo.properties.id, reference_id="main", path="test_obj", client=clt).upload(
        data=data, metadata={"foo": "bar"})

    copy_name = "copy_obj"
    copy = obj.copy("main", copy_name)
    obj_stat = obj.stat()
    copy_stat = copy.stat()

    assert copy != obj
    assert copy_stat.metadata == obj_stat.metadata
    assert copy_stat.path != obj_stat.path
    assert copy_stat.content_type == obj_stat.content_type
    assert copy_stat.physical_address != obj_stat.physical_address
    assert copy_stat.mtime >= obj_stat.mtime
    assert copy_stat.size_bytes == obj_stat.size_bytes
    assert copy_stat.checksum == obj_stat.checksum


def test_writer(setup_repo):
    _, repo = setup_repo
    obj = repo.branch("main").object("test_object")
    with obj.writer() as writer:
        with expect_exception_context(io.UnsupportedOperation):
            writer.seek(10)

        assert not writer.readable()
        assert writer.writable()

        writer.write("Hello")
        writer.write(" World!")

        # Check that the object does not exist in lakeFS before we close the writer
        with expect_exception_context(NotFoundException):
            obj.stat()

    assert obj.reader().read() == b"Hello World!"

    # try to close writer again
    writer.close()

    # write after close
    with expect_exception_context(ValueError, "I/O operation on closed file"):
        writer.write("test")

    # flush after close
    with expect_exception_context(ValueError, "I/O operation on closed file"):
        writer.flush()


@pytest.mark.parametrize("w_mode", get_args(WriteModes))
@pytest.mark.parametrize("r_mode", get_args(ReadModes))
@pytest.mark.parametrize("pre_sign", (True, False), indirect=True)
def test_writer_different_params(setup_repo, w_mode, r_mode, pre_sign):
    _, repo = setup_repo
    obj = repo.branch("main").object("test_object")
    writer = obj.writer(mode=w_mode)

    data = [
        "The quick brown fox jumps over the lazy dog"
        "Pack my box with five dozen liquor jugs"
        "Sphinx of black quartz, judge my vow" * 1000
    ]

    pos = 0
    for part in data:
        pos += len(part)
        writer.write(part)
        assert writer.tell() == pos

    # Check that the object does not exist in lakeFS before we close the writer
    with expect_exception_context(NotFoundException):
        obj.stat()

    writer.content_type = "text/plain"
    writer.metadata = {"foo": "bar"}
    writer.pre_sign = pre_sign

    writer.close()

    stats = obj.stat()
    assert stats.metadata == {"foo": "bar"}
    assert stats.content_type == "text/plain"
    assert stats.path == "test_object"
    assert stats.size_bytes == pos

    expected = "".join(data)
    res = obj.reader(mode=r_mode).read()

    if 'b' in r_mode:
        assert res.decode('utf-8') == expected
    else:
        assert res == expected


def _upload_file(repo, test_file):
    obj = repo.branch("main").object("test_obj")

    with open(test_file, "rb") as fd, obj.writer() as writer:
        writer.write(fd.read())

    return obj


def test_read_byte_by_byte(setup_repo):
    clt, repo = setup_repo

    data = b'test_data'
    obj = WriteableObject(repository_id=repo.properties.id, reference_id="main", path="test_obj", client=clt).upload(
        data=data, pre_sign=False)
    res = b""
    reader = obj.reader()
    while True:
        byte = reader.read(1)
        if not byte:
            break
        res += byte

    assert res == data


@TEST_DATA
def test_read_all(setup_repo, datafiles):
    _, repo = setup_repo
    test_file = datafiles / "data.csv"
    obj = _upload_file(repo, test_file)

    with open(test_file, "r", encoding="utf-8") as fd:
        data = fd.read()
        read_data = obj.reader("r").read()
        assert read_data == data


@TEST_DATA
def test_read_csv(setup_repo, datafiles):
    _, repo = setup_repo
    test_file = datafiles / "data.csv"
    obj = _upload_file(repo, test_file)

    uploaded = csv.reader(obj.reader('r'))

    with open(test_file, "r", encoding="utf-8") as fd:
        source = csv.reader(fd)
        for uploaded_row, source_row in zip(uploaded, source):
            assert uploaded_row == source_row


@TEST_DATA
def test_read_json(setup_repo, datafiles):
    _, repo = setup_repo
    test_file = datafiles / "data.json"
    obj = _upload_file(repo, test_file)

    with open(test_file, "r", encoding="utf-8") as fd:
        source = json.load(fd)
        uploaded = json.load(obj.reader(mode="r"))

        assert uploaded == source


@TEST_DATA
def test_read_yaml(setup_repo, datafiles):
    _, repo = setup_repo
    test_file = datafiles / "data.yaml"
    obj = _upload_file(repo, test_file)

    with open(test_file, "r", encoding="utf-8") as fd:
        source = yaml.load(fd, Loader=yaml.Loader)
        uploaded = yaml.load(obj.reader(mode="r"), Loader=yaml.Loader)
        assert uploaded == source


@TEST_DATA
def test_read_parquet(setup_repo, datafiles):
    _, repo = setup_repo
    test_file = datafiles / "data.parquet"
    obj = _upload_file(repo, test_file)

    uploaded = pd.read_parquet(obj.reader('rb'))
    print(uploaded)
    with open(test_file, "rb") as fd:
        source = pd.read_parquet(fd)
        assert source.equals(uploaded)


@TEST_DATA
def test_read_xml(setup_repo, datafiles):
    _, repo = setup_repo
    test_file = datafiles / "data.xml"
    obj = _upload_file(repo, test_file)

    uploaded = ElementTree.parse(obj.reader(mode="r"))
    source = ElementTree.parse(test_file)
    uploaded_root = uploaded.getroot()
    source_root = source.getroot()
    assert uploaded_root.tag == source_root.tag
    assert uploaded_root.attrib == source_root.attrib
    assert uploaded_root.text == source_root.text
    assert uploaded_root.tail == source_root.tail

    for uploaded_child, source_child in zip(uploaded_root, source_root):
        for uploaded_child_child, source_child_child in zip(uploaded_child, source_child):
            assert uploaded_child_child.tag == source_child_child.tag
            assert uploaded_child_child.attrib == source_child_child.attrib
            assert uploaded_child_child.text == source_child_child.text
            assert uploaded_child_child.tail == source_child_child.tail


def test_readline_no_newline(setup_repo):
    clt, repo = setup_repo
    data = b'test_data'
    obj = WriteableObject(repository_id=repo.properties.id, reference_id="main", path="test_obj", client=clt).upload(
        data=data, pre_sign=False)

    assert obj.reader().readline() == data

    assert obj.reader().readline() == data


def test_readline_partial_line_buffer(setup_repo):
    clt, repo = setup_repo
    data = "a" * 15 + "\n" + "b" * 25 + "\n" + "That is all folks! "
    obj = WriteableObject(repository_id=repo.properties.id, reference_id="main", path="test_obj", client=clt).upload(
        data=data, pre_sign=False)

    with obj.reader(mode="r") as reader:
        reader.seek(5)
        assert reader.readline() == "a" * 10 + "\n"
        assert reader.readline() == "b" * 25 + "\n"
        assert reader.readline() == "That is all folks! "
        assert reader.readline() == ""
        reader.seek(0)
        assert reader.readline() == "a" * 15 + "\n"
        assert reader.read() == data[16:]

    # Read with limit
    with obj.reader(mode="r") as reader:
        for _ in range(math.ceil(len(data) / 10)):
            end = 10
            index = data.find("\n")
            if -1 < index < 10:
                end = index + 1

            expected = data[:end]
            data = data[len(expected):]
            read = reader.readline(10)
            assert read == expected

        assert reader.read() == ""


def test_write_read_csv(setup_repo):
    _, repo = setup_repo
    columns = ["ID", "Name", "Email"]
    sample_data = [
        ['1', "Alice", "alice@example.com"],
        ['2', "Bob", "bob@example.com"],
        ['3', "Carol", "carol@example.com"],
    ]
    obj = repo.branch("main").object(path="csv/sample_data.csv")

    with obj.writer(mode='w', pre_sign=False, content_type="text/csv") as fd:
        writer = csv.writer(fd)
        writer.writerow(columns)
        for row in sample_data:
            writer.writerow(row)

    for i, row in enumerate(csv.reader(obj.reader(mode='r'))):
        if i == 0:
            assert row == columns
        else:
            assert row == sample_data[i - 1]


def test_reader_with_failure(setup_repo):
    _, repo = setup_repo
    obj = repo.branch("main").object("test_object")

    try:
        with obj.writer() as writer:
            writer.write("Hello World!")
            raise ValueError("Bad thing happened")
    except ValueError:
        pass

    # Verify writer is closed
    with expect_exception_context(ValueError, "I/O operation on closed file"):
        writer.write("test")

    # Check that the object does not exist in lakeFS after exception
    assert not obj.exists()


def test_writer_with_failure(setup_repo):
    clt, repo = setup_repo
    data = b"test_data"
    obj = WriteableObject(repository_id=repo.properties.id,
                          reference_id="main",
                          path="test_obj",
                          client=clt).upload(data=data)

    try:
        with obj.reader() as reader:
            reader.read()
            raise ValueError("Bad thing happened")
    except ValueError:
        pass

    assert reader.closed

    # Verify exception when trying to read is closed
    with expect_exception_context(ValueError, "I/O operation on closed file"):
        reader.read()


def test_writer_discard(setup_repo):
    _, repo = setup_repo
    obj = repo.branch("main").object("test_object")

    with obj.writer() as writer:
        writer.write("Hello World!")
        writer.discard()
        assert writer.closed

    assert not obj.exists()


@TEST_DATA
def test_writer_gif(setup_repo, datafiles):
    _, repo = setup_repo
    test_file = datafiles / "waving-axolotl.gif"
    image = Image.open(test_file)
    obj = repo.branch("main").object("waving-axolotl.gif")

    with obj.writer() as writer:
        image.save(writer, save_all=True, optimize=False, disposal=3)

    assert obj.exists()
    image.close()



"""
Module containing lakeFS reference implementation
"""

from __future__ import annotations

import base64
import binascii
import io
import json
import os
import tempfile
import urllib.parse
from abc import abstractmethod
from typing import AnyStr, IO, Iterator, List, Literal, Optional, Union, get_args

import lakefs_sdk
from lakefs_sdk import StagingMetadata

from lakefs.client import Client, _BaseLakeFSObject
from lakefs.exceptions import (
    api_exception_handler,
    handle_http_error,
    LakeFSException,
    NotFoundException,
    ObjectNotFoundException,
    NotAuthorizedException,
    ForbiddenException,
    PermissionException,
    ObjectExistsException,
    InvalidRangeException,
)
from lakefs.models import ObjectInfo

_LAKEFS_METADATA_PREFIX = "x-lakefs-meta-"
# _BUFFER_SIZE - Writer buffer size. While buffer size not exceed, data will be maintained in memory and file will
#                not be created.
_WRITER_BUFFER_SIZE = 32 * 1024 * 1024

ReadModes = Literal['r', 'rb']
WriteModes = Literal['x', 'xb', 'w', 'wb']
AllModes = Union[ReadModes, WriteModes]


class LakeFSIOBase(_BaseLakeFSObject, IO):
    """
    Base class for the lakeFS Reader and Writer classes
    """
    _obj: StoredObject
    _mode: AllModes
    _pos: int
    _pre_sign: Optional[bool] = None

    def __init__(self, obj: StoredObject, mode: AllModes, pre_sign: Optional[bool] = None,
                 client: Optional[Client] = None) -> None:
        self._obj = obj
        self._mode = mode
        self._pos = 0
        super().__init__(client)
        # must be set after super().__init__ to ensure the client is properly initialized.
        self._pre_sign = pre_sign if pre_sign is not None else self._client.storage_config.pre_sign_support

    @property
    def mode(self) -> str:
        """
        Returns the open mode for this object
        """
        return self._mode

    @property
    def name(self) -> str:
        """
        Returns the name of the object relative to the repo and reference
        """
        return self._obj.path

    def close(self) -> None:
        """
        Finalizes any existing operations on object and close open descriptors
        Inheriting classes need to override close() as needed.
        """

    @abstractmethod
    def _abort(self) -> None:
        """
        Specific implementation for when the IO object should be discarded
        """
        raise NotImplementedError

    def fileno(self) -> int:
        """
        The file descriptor number as defined by the operating system. In the context of lakeFS it has no meaning

        :raise io.UnsupportedOperation: Always, since fileno is not supported for lakeFS objects
        """
        raise io.UnsupportedOperation

    @abstractmethod
    def flush(self) -> None:
        raise NotImplementedError

    def isatty(self) -> bool:
        """
        Irrelevant for the lakeFS implementation
        """
        return False

    @abstractmethod
    def readable(self) -> bool:
        raise NotImplementedError

    def readline(self, limit: int = -1):
        """
        Must be explicitly implemented by inheriting class
        """
        raise io.UnsupportedOperation

    def readlines(self, hint: int = -1):
        """
        Must be explicitly implemented by inheriting class
        """
        raise io.UnsupportedOperation

    @abstractmethod
    def seekable(self) -> bool:
        raise NotImplementedError

    def truncate(self, size: int = None) -> int:
        """
        Unsupported by lakeFS implementation
        """
        raise io.UnsupportedOperation

    @abstractmethod
    def writable(self) -> bool:
        raise NotImplementedError

    @abstractmethod
    def write(self, s: AnyStr) -> int:
        raise NotImplementedError

    def writelines(self, lines: List[AnyStr]) -> None:
        """
        Unsupported by lakeFS implementation
        """
        raise io.UnsupportedOperation

    def __next__(self) -> AnyStr:
        line = self.readline()
        if len(line) == 0:
            raise StopIteration
        return line

    def __iter__(self) -> Iterator[AnyStr]:
        return self

    def __enter__(self) -> LakeFSIOBase:
        return self

    def __exit__(self, typ, value, traceback) -> bool:
        if typ is None:  # Perform logic in case no exception was raised
            self.close()

        else:
            self._abort()  # perform logic regardless of exception

        return False  # Don't suppress an exception

    @abstractmethod
    def seek(self, offset: int, whence: int = 0) -> int:
        raise NotImplementedError

    @abstractmethod
    def read(self, n: int = None) -> str | bytes:
        raise NotImplementedError

    def tell(self) -> int:
        """
        For readers - read position, for writers can be used as an indication for bytes written
        """
        return self._pos


class ObjectReader(LakeFSIOBase):
    """
    ObjectReader provides read-only functionality for lakeFS objects with IO semantics.
    This Object is instantiated and returned for immutable reference types (Commit, Tag...)
    """
    _readlines_buf: io.BytesIO

    def __init__(self, obj: StoredObject, mode: ReadModes, pre_sign: Optional[bool] = None,
                 client: Optional[Client] = None) -> None:
        if mode not in get_args(ReadModes):
            raise ValueError(f"invalid read mode: '{mode}'. ReadModes: {ReadModes}")

        super().__init__(obj, mode, pre_sign, client)
        self._readlines_buf = io.BytesIO(b"")
        self._is_closed = False

    @property
    def pre_sign(self):
        """
        Returns whether the pre_sign mode is enabled
        """
        if self._pre_sign is None:
            self._pre_sign = self._client.storage_config.pre_sign_support
        return self._pre_sign

    @pre_sign.setter
    def pre_sign(self, value: bool) -> None:
        """
        Set the pre_sign mode to value

        :param value: The new value for pre_sign mode
        """
        self._pre_sign = value

    @property
    def closed(self) -> bool:
        """
        Returns True after the object is closed
        """
        return self._is_closed

    def readable(self) -> bool:
        """
        Returns True always
        """
        return True

    def write(self, s: AnyStr) -> int:
        """
        Unsupported for reader object
        """
        raise io.UnsupportedOperation

    def seekable(self) -> bool:
        """
        Returns True always
        """
        return True

    def writable(self) -> bool:
        """
        Unsupported - read only object
        """
        return False

    def seek(self, offset: int, whence: int = 0) -> int:
        """
        Move the object's reading position

        :param offset: The offset from the beginning of the file
        :param whence: Optional. The whence argument is optional and defaults to
            os.SEEK_SET or 0 (absolute file positioning);
            other values are os.SEEK_CUR or 1 (seek relative to the current position) and os.SEEK_END or 2
            (seek relative to the file’s end)
        :raise ValueError: if reader is closed
        :raise OSError: if calculated new position is negative
        :raise io.UnsupportedOperation: If whence value is unsupported
        """
        if self._is_closed:
            raise ValueError("I/O operation on closed file")

        if whence == os.SEEK_SET:
            pos = offset
        elif whence == os.SEEK_CUR:
            pos = self._pos + offset
        elif whence == os.SEEK_END:
            size = self._obj.stat().size_bytes  # Seek end requires us to know the size of the file
            pos = size + offset
        else:
            raise io.UnsupportedOperation(f"whence={whence} is not supported")

        if pos < 0:
            raise OSError("position must be a non-negative integer")
        self._pos = pos
        return pos

    def _cast_by_mode(self, retval):
        if 'b' not in self.mode:
            return retval.decode('utf-8')
        return retval

    def _read(self, read_range: str) -> str | bytes:
        try:
            with api_exception_handler(_io_exception_handler):
                return self._client.sdk_client.objects_api.get_object(self._obj.repo,
                                                                      self._obj.ref,
                                                                      self._obj.path,
                                                                      range=read_range,
                                                                      presign=self.pre_sign)

        except InvalidRangeException:
            # This is done in order to behave like the built-in open() function
            return b''

    def read(self, n: int = None) -> str | bytes:
        """
        Read object data

        :param n: How many bytes to read. If read_bytes is None, will read from current position to end.
            If current position + read_bytes > object size.
        :return: The bytes read
        :raise ValueError: if reader is closed
        :raise OSError: if read_bytes is non-positive
        :raise ObjectNotFoundException: if repository id, reference id or object path does not exist
        :raise PermissionException: if user is not authorized to perform this operation, or operation is forbidden
        :raise ServerException: for any other errors
        """
        if self._is_closed:
            raise ValueError("I/O operation on closed file")

        if n and n <= 0:
            raise OSError("read_bytes must be a positive integer")

        read_range = self._get_range_string(start=self._pos, read_bytes=n)
        contents = self._read(read_range)
        self._pos += len(contents)  # Update pointer position

        return self._cast_by_mode(contents)

    def readline(self, limit: int = -1):
        """
        Read and return a line from the stream.

        :param limit: If limit > -1 returns at most limit bytes
        :raise ValueError: if reader is closed
        :raise ObjectNotFoundException: if repository id, reference id or object path does not exist
        :raise PermissionException: if user is not authorized to perform this operation, or operation is forbidden
        :raise ServerException: for any other errors
        """
        if self._is_closed:
            raise ValueError("I/O operation on closed file")

        if self._readlines_buf.getbuffer().nbytes == 0:
            self._readlines_buf = io.BytesIO(self._read(self._get_range_string(0)))
        self._readlines_buf.seek(self._pos)
        line = self._readlines_buf.readline(limit)
        self._pos = self._readlines_buf.tell()
        return self._cast_by_mode(line)

    def flush(self) -> None:
        """
        Nothing to do for reader

        :raise ValueError: if reader is closed
        """
        if self._is_closed:
            raise ValueError("I/O operation on closed file")

    def close(self) -> None:
        """
        Close open descriptors
        """
        if self._is_closed:
            return

        self._is_closed = True
        self._readlines_buf.close()

    def _abort(self):
        """
        Closes reader
        """
        self.close()

    @staticmethod
    def _get_range_string(start, read_bytes=None):
        if start == 0 and read_bytes is None:
            return None
        if read_bytes is None:
            return f"bytes={start}-"
        return f"bytes={start}-{start + read_bytes - 1}"

    def __str__(self):
        return self._obj.path

    def __repr__(self):
        return f'ObjectReader(path="{self._obj.path}")'


class ObjectWriter(LakeFSIOBase):
    """
    ObjectWriter provides write-only functionality for lakeFS objects with IO semantics.
    This Object is instantiated and returned from the WriteableObject writer method.
    For the data to be actually written to the lakeFS server the close() method must be invoked explicitly or
    implicitly when using writer as a context.
    """
    _fd: tempfile.SpooledTemporaryFile
    _obj_stats: ObjectInfo = None

    def __init__(self,
                 obj: StoredObject,
                 mode: WriteModes,
                 pre_sign: Optional[bool] = None,
                 content_type: Optional[str] = None,
                 metadata: Optional[dict[str, str]] = None,
                 client: Optional[Client] = None) -> None:

        if 'x' in mode and obj.exists():  # Requires explicit create
            raise ObjectExistsException

        if mode not in get_args(WriteModes):
            raise ValueError(f"invalid write mode: '{mode}'. WriteModes: {WriteModes}")

        self.content_type = content_type
        self.metadata = metadata

        open_kwargs = {
            "encoding": "utf-8" if 'b' not in mode else None,
            "mode": 'b+',  # Always write to file in binary mode (bug in urllib3 < 2.0,
            "max_size": _WRITER_BUFFER_SIZE,
        }
        self._fd = tempfile.SpooledTemporaryFile(**open_kwargs)  # pylint: disable=consider-using-with
        super().__init__(obj, mode, pre_sign, client)

    @property
    def pre_sign(self) -> bool:
        """
        Returns whether the pre_sign mode is enabled
        """
        if self._pre_sign is None:
            self._pre_sign = self._client.storage_config.pre_sign_support
        return self._pre_sign

    @pre_sign.setter
    def pre_sign(self, value: bool) -> None:
        """
        Set the pre_sign mode to value

        :param value: The new value for pre_sign mode
        """
        self._pre_sign = value

    @property
    def closed(self) -> bool:
        """
        Returns True after the object is closed
        """
        return self._fd.closed

    def flush(self) -> None:
        """
        Flush buffer to file. Prevent flush if total write size is still smaller than _BUFFER_SIZE so that we avoid
        unnecessary write to disk.

        :raise ValueError: if writer is closed
        """

        if self._fd.closed:
            raise ValueError("I/O operation on closed file")

        # Don't flush buffer to file if we didn't exceed buffer size
        # We want to avoid using the file if possible
        if self._pos > _WRITER_BUFFER_SIZE:
            self._fd.flush()

    def write(self, s: AnyStr) -> int:
        """
        Write data to buffer

        :param s: The data to write
        :return: The number of bytes written to buffer
        :raise ValueError: if writer is closed
        """
        contents = s.encode('utf-8') if isinstance(s, str) else s
        count = self._fd.write(contents)
        self._pos += count
        return count

    def discard(self) -> None:
        """
        Discards of the write buffer and closes writer
        """
        self._abort()

    def close(self) -> None:
        """
        Write the data to the lakeFS server
        """
        if self._fd.closed:
            return

        stats = self._upload_presign() if self.pre_sign else self._upload_raw()
        self._obj_stats = ObjectInfo(**stats.dict())
        self._fd.close()

    def _abort(self) -> None:
        """
        Close open descriptors but create nothing on lakeFS.
        """
        if not self._fd.closed:
            self._fd.close()

    @staticmethod
    def _extract_etag_from_response(headers) -> str:
        # prefer Content-MD5 if exists
        content_md5 = headers.get("Content-MD5")
        if content_md5 is not None and len(content_md5) > 0:
            try:  # decode base64, return as hex
                decode_md5 = base64.b64decode(content_md5)
                return binascii.hexlify(decode_md5).decode("utf-8")
            except binascii.Error:
                pass

        # fallback to ETag
        etag = headers.get("ETag", "").strip(' "')
        return etag

    def _upload_raw(self) -> lakefs_sdk.ObjectStats:
        """
        Use raw upload API call to bypass validation of content parameter
        """
        auth_settings = ['basic_auth', 'cookie_auth', 'oidc_auth', 'saml_auth', 'jwt_token']
        headers = {
            "Accept": "application/json",
            "Content-Type": self.content_type if self.content_type is not None else "application/octet-stream"
        }

        # Create user metadata headers
        if self.metadata is not None:
            for k, v in self.metadata.items():
                headers[_LAKEFS_METADATA_PREFIX + k] = v

        self._fd.seek(0)
        resource_path = urllib.parse.quote(f"/repositories/{self._obj.repo}/branches/{self._obj.ref}/objects",
                                           encoding="utf-8")
        query_params = urllib.parse.urlencode({"path": self._obj.path}, encoding="utf-8")
        url = self._client.config.host + resource_path + f"?{query_params}"
        self._client.sdk_client.objects_api.api_client.update_params_for_auth(headers, None, auth_settings,
                                                                              resource_path, "POST", self._fd)
        resp = self._client.sdk_client.objects_api.api_client.rest_client.pool_manager.request(url=url,
                                                                                               method="POST",
                                                                                               headers=headers,
                                                                                               body=self._fd)

        handle_http_error(resp)
        return lakefs_sdk.ObjectStats(**json.loads(resp.data))

    def _upload_presign(self) -> lakefs_sdk.ObjectStats:
        staging_location = self._client.sdk_client.staging_api.get_physical_address(self._obj.repo,
                                                                                    self._obj.ref,
                                                                                    self._obj.path,
                                                                                    True)
        url = staging_location.presigned_url

        headers = {"Content-Length": self._pos}
        if self.content_type:
            headers["Content-Type"] = self.content_type
        if self._client.storage_config.blockstore_type == "azure":
            headers["x-ms-blob-type"] = "BlockBlob"

        self._fd.seek(0)
        resp = self._client.sdk_client.staging_api.api_client.rest_client.pool_manager.request(method="PUT",
                                                                                               url=url,
                                                                                               body=self._fd,
                                                                                               headers=headers)
        handle_http_error(resp)

        etag = ObjectWriter._extract_etag_from_response(resp.headers)
        size_bytes = self._pos
        staging_metadata = StagingMetadata(staging=staging_location,
                                           size_bytes=size_bytes,
                                           checksum=etag,
                                           user_metadata=self.metadata,
                                           content_type=self.content_type)
        return self._client.sdk_client.staging_api.link_physical_address(self._obj.repo,
                                                                         self._obj.ref,
                                                                         self._obj.path,
                                                                         staging_metadata=staging_metadata)

    def readable(self) -> bool:
        """
        ObjectWriter is write-only - return False always
        """
        return False

    def seekable(self) -> bool:
        """
        ObjectWriter is not seekable. Returns False always
        """
        return False

    def writable(self) -> bool:
        """
        Returns True always
        """
        return True

    def seek(self, offset: int, whence: int = 0) -> int:
        """
        Unsupported for writer class
        """
        raise io.UnsupportedOperation

    def read(self, n: int = None) -> str | bytes:
        """
        Unsupported for writer class
        """
        raise io.UnsupportedOperation

    def __repr__(self):
        return f'ObjectWriter(path="{self._obj.path}")'


class StoredObject(_BaseLakeFSObject):
    """
    Class representing an object in lakeFS.
    """
    _repo_id: str
    _ref_id: str
    _path: str
    _stats: Optional[ObjectInfo] = None

    def __init__(self, repository_id: str, reference_id: str, path: str, client: Optional[Client] = None):
        self._repo_id = repository_id
        self._ref_id = reference_id
        self._path = path
        super().__init__(client)

    def __str__(self) -> str:
        return self.path

    def __repr__(self):
        return f'StoredObject(repository="{self.repo}", reference="{self.ref}", path="{self.path}")'

    @property
    def repo(self) -> str:
        """
        Returns the object's repository id
        """
        return self._repo_id

    @property
    def ref(self) -> str:
        """
        Returns the object's reference id
        """
        return self._ref_id

    @property
    def path(self) -> str:
        """
        Returns the object's path relative to repository and reference ids
        """
        return self._path

    def reader(self, mode: ReadModes = 'rb', pre_sign: Optional[bool] = None) -> ObjectReader:
        """
        Context manager which provide a file-descriptor like object that allow reading the given object.

        Usage Example:

        .. code-block:: python

            import lakefs

            obj = lakefs.repository("<repository_name>").branch("<branch_name>").object("file.txt")
            file_size = obj.stat().size_bytes

            with obj.reader(mode='r', pre_sign=True) as fd:
                # print every other 10 chars
                while fd.tell() < file_size
                    print(fd.read(10))
                    fd.seek(10, os.SEEK_CUR)

        :param mode: Read mode - as supported by ReadModes
        :param pre_sign: (Optional), enforce the pre_sign mode on the lakeFS server. If not set, will probe server for
            information.
        :return: A Reader object
        """
        return ObjectReader(self, mode=mode, pre_sign=pre_sign, client=self._client)

    def stat(self) -> ObjectInfo:
        """
        Return the Stat object representing this object
        """
        if self._stats is None:
            with api_exception_handler(_io_exception_handler):
                stat = self._client.sdk_client.objects_api.stat_object(self._repo_id, self._ref_id, self._path)
                self._stats = ObjectInfo(**stat.dict())
        return self._stats

    def exists(self) -> bool:
        """
        Returns True if object exists in lakeFS, False otherwise
        """

        exists = False

        def exist_handler(e: LakeFSException):
            if isinstance(e, NotFoundException):
                return None  # exists = False
            return _io_exception_handler(e)

        with api_exception_handler(exist_handler):
            self._client.sdk_client.objects_api.head_object(self._repo_id, self._ref_id, self._path)
            exists = True

        return exists

    def copy(self, destination_branch_id: str, destination_path: str) -> WriteableObject:
        """
        Copy the object to a destination branch

        :param destination_branch_id: The destination branch to copy the object to
        :param destination_path: The path of the copied object in the destination branch
        :return: The newly copied Object
        :raise ObjectNotFoundException: if repo id,reference id, destination branch id or object path does not exist
        :raise PermissionException: if user is not authorized to perform this operation, or operation is forbidden
        :raise ServerException: for any other errors
        """

        with api_exception_handler():
            object_copy_creation = lakefs_sdk.ObjectCopyCreation(src_ref=self._ref_id, src_path=self._path)
            self._client.sdk_client.objects_api.copy_object(repository=self._repo_id,
                                                            branch=destination_branch_id,
                                                            dest_path=destination_path,
                                                            object_copy_creation=object_copy_creation)

        return WriteableObject(repository_id=self._repo_id, reference_id=destination_branch_id, path=destination_path,
                               client=self._client)


class WriteableObject(StoredObject):
    """
    WriteableObject inherits from ReadableObject and provides read/write functionality for lakeFS objects
    using IO semantics.
    This Object is instantiated and returned upon invoking writer() on Branch reference type.
    """

    def __init__(self, repository_id: str, reference_id: str, path: str,
                 client: Optional[Client] = None) -> None:
        super().__init__(repository_id, reference_id, path, client=client)

    def __repr__(self):
        return f'WriteableObject(repository="{self.repo}", reference="{self.ref}", path="{self.path}")'

    def upload(self,
               data: str | bytes,
               mode: WriteModes = 'w',
               pre_sign: Optional[bool] = None,
               content_type: Optional[str] = None,
               metadata: Optional[dict[str, str]] = None) -> WriteableObject:
        """
        Upload a new object or overwrites an existing object

        :param data: The contents of the object to write (can be bytes or string)
        :param mode: Write mode:

            'x'     - Open for exclusive creation

            'xb'    - Open for exclusive creation in binary mode

            'w'     - Create a new object or truncate if exists

            'wb'    - Create or truncate in binary mode
        :param pre_sign: (Optional) Explicitly state whether to use pre_sign mode when uploading the object.
            If None, will be taken from pre_sign property.
        :param content_type: (Optional) Explicitly set the object Content-Type
        :param metadata: (Optional) User metadata
        :return: The Stat object representing the newly created object
        :raise ObjectExistsException: if object exists and mode is exclusive ('x')
        :raise ObjectNotFoundException: if repo id, reference id or object path does not exist
        :raise PermissionException: if user is not authorized to perform this operation, or operation is forbidden
        :raise ServerException: for any other errors
        """
        with ObjectWriter(self, mode, pre_sign, content_type, metadata, self._client) as writer:
            writer.write(data)

        return self

    def delete(self) -> None:
        """
        Delete object from lakeFS

        :raise ObjectNotFoundException: if repo id, reference id or object path does not exist
        :raise PermissionException: if user is not authorized to perform this operation, or operation is forbidden
        :raise ServerException: for any other errors
        """
        with api_exception_handler(_io_exception_handler):
            self._client.sdk_client.objects_api.delete_object(self._repo_id, self._ref_id, self._path)
            self._stats = None

    def writer(self,
               mode: WriteModes = 'wb',
               pre_sign: Optional[bool] = None,
               content_type: Optional[str] = None,
               metadata: Optional[dict[str, str]] = None) -> ObjectWriter:
        """
        Context manager which provide a file-descriptor like object that allow writing the given object to lakeFS
        The writes are saved in a buffer as long as the writer is open. Only when it closes it writes the data into
        lakeFS. The optional parameters can be modified by accessing the respective fields as long as the writer is
        still open.

        Usage example of reading a file from local file system and writing it to lakeFS:

        .. code-block:: python

            import lakefs

            obj = lakefs.repository("<repository_name>").branch("<branch_name>").object("my_image")

            with open("my_local_image", mode='rb') as reader, obj.writer("wb") as writer:
                writer.write(reader.read())

        :param mode: Write mode - as supported by WriteModes
        :param pre_sign: (Optional), enforce the pre_sign mode on the lakeFS server. If not set, will probe server for
            information.
        :param content_type: (Optional) Specify the data media type
        :param metadata: (Optional) User defined metadata to save on the object
        :return: A Writer object
        """
        return ObjectWriter(self,
                            mode=mode,
                            pre_sign=pre_sign,
                            content_type=content_type,
                            metadata=metadata,
                            client=self._client)


def _io_exception_handler(e: LakeFSException):
    if isinstance(e, NotFoundException):
        return ObjectNotFoundException(e.status_code, e.reason)
    if isinstance(e, (NotAuthorizedException, ForbiddenException)):
        return PermissionException(e.status_code, e.reason)
    return e



import time
from datetime import timedelta

import lakefs_sdk

from lakefs.exceptions import ImportManagerException
from lakefs.import_manager import ImportManager
from tests.utests.common import get_test_client, expect_exception_context


class TestImportManager:
    def test_import_start_wait(self, monkeypatch):
        clt = get_test_client()
        mgr = ImportManager("test_repo", "test_branch", client=clt)

        # Wait before start
        with expect_exception_context(ImportManagerException):
            mgr.wait()

        with monkeypatch.context():
            def monkey_import_start(*_, **__):
                return lakefs_sdk.ImportCreationResponse(id="import_id")

            monkeypatch.setattr(lakefs_sdk.ImportApi, "import_start", monkey_import_start)
            res = mgr.start()
            assert res == "import_id"

            # try again and expect import in progress
            with expect_exception_context(ImportManagerException):
                mgr.start()

            requests = 5
            status = lakefs_sdk.ImportStatus(completed=False,
                                             update_time=time.time(),
                                             ingested_objects=500,
                                             metarange_id=None,
                                             commit=None,
                                             error=None)

            def monkey_import_status(*_, **__):
                nonlocal requests, status
                requests -= 1
                if requests == 0:
                    status.completed = True
                return status

            monkeypatch.setattr(lakefs_sdk.ImportApi, "import_status", monkey_import_status)
            res = mgr.wait(poll_interval=timedelta())
            assert res.completed
            assert requests == 0

            # try again and expect no error
            mgr.wait()



import pytest

import lakefs
from lakefs.exceptions import BadRequestException


def test_reference_log(setup_branch_with_commits):
    branch = setup_branch_with_commits

    for i, c in enumerate(branch.log(max_amount=199)):
        assert c.message == f"commit {i}"

    commits = list(branch.log(max_amount=2000))
    for i, c in enumerate(commits[:-1]):  # Ignore initial commit
        assert c.message == f"commit {i}"
    assert len(commits) == 200

    assert len(list(branch.log(limit=True, amount=10, max_amount=100))) == 10


def test_reference_diff(setup_branch_with_commits):
    branch = setup_branch_with_commits

    commits = list(branch.log(max_amount=2))
    assert len(list(branch.diff(branch.get_commit().id))) == 0
    changes = list(branch.diff(commits[0].id, type="two_dot"))
    assert len(changes) == 0

    changes = list(branch.diff(commits[1].id, type="two_dot"))
    assert len(changes) == 1
    assert changes[0].path == "test1"
    assert changes[0].type == "removed"

    other_branch = lakefs.Repository(branch.repo_id).branch("other_branch").create("test_branch")
    other_branch.object("prefix1/test1").upload(data="data1")
    other_branch.object("prefix2/test2").upload(data="data2")
    other_branch.commit("other commit")

    changes = list(branch.diff(other_branch))
    assert len(changes) == 2

    changes = list(branch.diff(other_branch, prefix="prefix2"))
    assert len(changes) == 1
    assert changes[0].path == "prefix2/test2"
    assert changes[0].type == "added"


def test_reference_merge_into(setup_branch_with_commits):
    branch = setup_branch_with_commits
    repo = lakefs.Repository(branch.repo_id)
    main = repo.branch("main")

    # test merging into same branch
    branch_a = repo.branch("test_branch_merge_into_a").create(branch)
    branch_b = repo.branch("test_branch_merge_into_b").create(branch)

    with pytest.raises(BadRequestException, match=r'.+no changes.+'):
        branch_a.merge_into(branch_b, message="MergeNoChanges")

    branch_a.merge_into(branch_b, message="MergeNoChangesWithFlag", allow_empty=True)

    # test merging into other branch
    commits = list(branch.log(max_amount=2))
    other_branch = repo.branch("test_reference_merge_into").create(main)
    ref = repo.ref(commits[1].id)
    ref.merge_into(other_branch, message="Merge1")
    assert other_branch.get_commit().message == "Merge1"
    assert list(other_branch.log(max_amount=2))[1].id == commits[1].id

    branch.merge_into(other_branch.id, message="Merge2")
    assert other_branch.get_commit().message == "Merge2"
    assert list(other_branch.log(max_amount=3))[2].id == commits[0].id


def test_reference_objects(setup_repo):
    _, repo = setup_repo
    test_branch = repo.branch("main")
    path_and_data = ["a", "b", "bar/a", "bar/b", "bar/c", "c", "foo/a", "foo/b", "foo/c"]
    for s in path_and_data:
        test_branch.object(s).upload(s)

    objects = list(test_branch.objects())
    assert len(objects) == len(path_and_data)
    for obj in objects:
        assert obj.path in path_and_data

    expected = ["a", "b", "bar/", "c", "foo/"]
    i = 0
    for obj in test_branch.objects(delimiter='/'):
        i += 1
        assert obj.path in expected

    assert i == len(expected)



"""
LenientNamedTuple Module
"""


class LenientNamedTuple:
    """
    Class which provides the NamedTuple functionality but allows initializing with unknown fields.
    The unknown fields will be ignored and mandatory fields are enforced
    """

    __initialized: bool = False
    unknown: dict = {}

    def __init__(self, **kwargs):
        fields = list(self.__class__.__dict__["__annotations__"].keys())
        for k, v in kwargs.items():
            if k in fields:
                setattr(self, k, v)
                fields.remove(k)
            else:
                self.unknown[k] = v

        if len(fields) > 0:
            raise TypeError(f"missing {len(fields)} required arguments: {fields}")

        self.__initialized = True
        super().__init__()

    def __repr__(self):
        class_name = self.__class__.__name__
        if hasattr(self, 'id'):
            return f'{class_name}(id="{self.id}")'
        return f'{class_name}()'

    def __setattr__(self, name, value):
        if self.__initialized:
            raise AttributeError("can't set attribute")
        super().__setattr__(name, value)

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        for k, v in self.__dict__.items():
            if k == "unknown":
                continue
            if other.__dict__[k] != v:
                return False

        return True

    def __str__(self):
        fields = {}
        for k, v in self.__dict__.items():
            if k != "unknown" and k[0] != "_":  # Filter internal and unknown fields
                fields[k] = v
        return str(fields)



from time import sleep

import pytest

from lakefs.client import Client
from lakefs.exceptions import ImportManagerException, ConflictException
from tests.utests.common import expect_exception_context

_IMPORT_PATH = "s3://esti-system-testing-data/import-test-data/"

_FILES_TO_CHECK = ["nested/prefix-1/file002005",
                   "nested/prefix-2/file001894",
                   "nested/prefix-3/file000005",
                   "nested/prefix-4/file000645",
                   "nested/prefix-5/file001566",
                   "nested/prefix-6/file002011",
                   "nested/prefix-7/file000101", ]


def skip_on_unsupported_blockstore(clt: Client, supported_blockstores: [str]):
    if clt.storage_config.blockstore_type not in supported_blockstores:
        pytest.skip(f"Unsupported blockstore type for test: {clt.storage_config.blockstore_type}")


def test_import_manager(setup_repo):
    clt, repo = setup_repo
    skip_on_unsupported_blockstore(clt, "s3")
    branch = repo.branch("import-branch").create("main")
    mgr = branch.import_data(commit_message="my imported data", metadata={"foo": "bar"})

    #  No import running
    with expect_exception_context(ImportManagerException):
        mgr.cancel()

    # empty import
    res = mgr.run()
    assert res.error is None
    assert res.completed
    assert res.commit.id == branch.get_commit().id
    assert res.commit.message == "my imported data"
    assert res.commit.metadata.get("foo") == "bar"
    assert res.ingested_objects == 0

    # Expect failure trying to run manager twice
    with expect_exception_context(ImportManagerException):
        mgr.run()

    # Import with objects and prefixes
    mgr = branch.import_data()
    dest_prefix = "imported/new-prefix/"
    mgr.prefix(_IMPORT_PATH + "prefix-1/",
               dest_prefix + "prefix-1/").prefix(_IMPORT_PATH + "prefix-2/",
                                                 dest_prefix + "prefix-2/")
    for o in _FILES_TO_CHECK:
        mgr.object(_IMPORT_PATH + o, dest_prefix + o)
    mgr.commit_message = "new commit"
    mgr.commit_metadata = None
    res = mgr.run()

    assert res.error is None
    assert res.completed
    assert res.commit.id == branch.get_commit().id
    assert res.commit.message == mgr.commit_message
    assert res.commit.metadata.get("foo") is None
    assert res.ingested_objects == 4207

    # Conflict since import completed
    with expect_exception_context(ConflictException):
        mgr.cancel()


def test_import_manager_cancel(setup_repo):
    clt, repo = setup_repo
    skip_on_unsupported_blockstore(clt, "s3")
    branch = repo.branch("import-branch").create("main")
    expected_commit_id = branch.get_commit().id
    expected_commit_message = branch.get_commit().message

    mgr = branch.import_data(commit_message="my imported data", metadata={"foo": "bar"})
    mgr.prefix(_IMPORT_PATH, "import/")

    mgr.start()
    sleep(1)

    with expect_exception_context(ImportManagerException):
        mgr.start()

    mgr.cancel()

    status = mgr.status()
    assert branch.get_commit().id == expected_commit_id
    assert branch.get_commit().message == expected_commit_message
    assert not status.completed
    assert "Canceled" in status.error.message
    assert len(mgr.sources) == 1



"""
Import module provides a simpler interface to the lakeFS SDK import functionality
"""

from __future__ import annotations

import asyncio
from datetime import timedelta
from typing import Optional, Dict, List

import lakefs_sdk

from lakefs.models import ImportStatus, _OBJECT, _COMMON_PREFIX
from lakefs.client import Client, _BaseLakeFSObject
from lakefs.exceptions import ImportManagerException, api_exception_handler


class ImportManager(_BaseLakeFSObject):
    """
    ImportManager provides an easy-to-use interface to perform imports with multiple sources.
    It provides both synchronous and asynchronous functionality allowing the user to start an import process,
    continue executing logic and poll for the import completion.

    ImportManager usage example:

    .. code-block:: python

        import lakefs

        branch = lakefs.repository("<repository_name>").branch("<branch_name>")
        mgr = branch.import_data(commit_message="my imported data", metadata={"foo": "bar"})

        # add sources for import
        mgr.prefix(object_store_uri="s3://import-bucket/data1/",
                   destination="import-prefix/").object(object_store_uri="s3://import-bucket/data2/imported_file",
                                                        destination="import-prefix/imported_file")
        # start import and wait
        mgr.run()

    """
    _repo_id: str
    _branch_id: str
    _in_progress: bool = False
    _import_id: str = None
    commit_message: str
    commit_metadata: Optional[Dict]
    sources: List[lakefs_sdk.ImportLocation]

    def __init__(self, repository_id: str, branch_id: str, commit_message: str = "",
                 commit_metadata: Optional[Dict] = None, client: Optional[Client] = None) -> None:
        self._repo_id = repository_id
        self._branch_id = branch_id
        self.commit_message = commit_message
        self.commit_metadata = commit_metadata
        self.sources = []
        super().__init__(client)

    @property
    def import_id(self) -> str:
        """
        Returns the id of the current import process
        """
        return self._import_id

    def _append_source(self, import_location: lakefs_sdk.ImportLocation):
        if self._import_id is not None:
            raise ImportManagerException("Cannot add additional sources to an already started import")

        self.sources.append(import_location)

    def prefix(self, object_store_uri: str, destination: str) -> ImportManager:
        """
        Creates a new import source of type "common_prefix" and adds it to the list of sources

        :param object_store_uri: The URI from which to import the objects
        :param destination: The destination prefix relative to the branch
        :return: The ImportManager instance (self) after update, to allow operations chaining
        """
        self._append_source(lakefs_sdk.ImportLocation(type=_COMMON_PREFIX,
                                                      path=object_store_uri,
                                                      destination=destination))
        return self

    def object(self, object_store_uri: str, destination: str) -> ImportManager:
        """
        Creates a new import source of type "object" and adds it to the list of sources

        :param object_store_uri: The URI from which to import the object
        :param destination: The destination path for the object relative to the branch
        :return: The ImportManager instance (self) after update, to allow operations chaining
        """
        self._append_source(lakefs_sdk.ImportLocation(type=_OBJECT, path=object_store_uri, destination=destination))
        return self

    def start(self) -> str:
        """
        Start import, reporting back (and storing) a process id

        :return: The import process identifier in lakeFS
        :raise ImportManagerException: if an import process is already in progress
        :raise NotFoundException: if branch or repository do not exist
        :raise NotAuthorizedException: if user is not authorized to perform this operation
        :raise ValidationError: if path_type is not one of the allowed values
        :raise ServerException: for any other errors
        """
        if self._in_progress:
            raise ImportManagerException("Import in progress")
        if self._import_id is not None:
            raise ImportManagerException("Import Manager can only be used once")

        creation = lakefs_sdk.ImportCreation(paths=self.sources,
                                             commit=lakefs_sdk.CommitCreation(message=self.commit_message,
                                                                              metadata=self.commit_metadata))
        with api_exception_handler():
            res = self._client.sdk_client.import_api.import_start(repository=self._repo_id,
                                                                  branch=self._branch_id,
                                                                  import_creation=creation)
            self._import_id = res.id
            self._in_progress = True

        return self._import_id

    async def _wait_for_completion(self, poll_interval: timedelta) -> lakefs_sdk.ImportStatus:
        while True:
            await asyncio.sleep(poll_interval.total_seconds())
            with api_exception_handler():
                resp = self._client.sdk_client.import_api.import_status(repository=self._repo_id,
                                                                        branch=self._branch_id,
                                                                        id=self._import_id)
            if resp.completed:
                return resp
            if resp.error is not None:
                raise ImportManagerException(f"Import Error: {resp.error.message}")

    def wait(self, poll_interval: Optional[timedelta] = timedelta(seconds=2)) -> ImportStatus:
        """
        Poll a started import task ID, blocking until completion

        :param poll_interval: The interval for polling the import status.
        :return: Import status as returned by the lakeFS server
        :raise ImportManagerException: if no import is in progress
        :raise NotFoundException: if branch, repository or import id do not exist
        :raise NotAuthorizedException: if user is not authorized to perform this operation
        :raise ServerException: for any other errors
        """
        if self._import_id is None:
            raise ImportManagerException("No import in progress")

        res = asyncio.run(self._wait_for_completion(poll_interval))
        self._in_progress = False
        return ImportStatus(**res.dict())

    def run(self, poll_interval: Optional[timedelta] = None) -> ImportStatus:
        """
        Same as calling start() and then wait()

        :param poll_interval: The interval for polling the import status.
        :return: Import status as returned by the lakeFS server
        :raises: See start(), wait()
        """
        self.start()
        wait_kwargs = {} if poll_interval is None else {"poll_interval": poll_interval}
        return self.wait(**wait_kwargs)

    def cancel(self) -> None:
        """
        Cancel an ongoing import process

        :raise NotFoundException: if branch, repository or import id do not exist
        :raise NotAuthorizedException: if user is not authorized to perform this operation
        :raise ConflictException: if the import was already completed
        :raise ServerException: for any other errors
        """
        if self._import_id is None:  # Can't cancel on no id
            raise ImportManagerException("No import in progress")

        with api_exception_handler():
            self._client.sdk_client.import_api.import_cancel(repository=self._repo_id,
                                                             branch=self._branch_id,
                                                             id=self._import_id)
            self._in_progress = False

    def status(self) -> ImportStatus:
        """
        Get the current import status

        :return: Import status as returned by the lakeFS server
        :raise ImportManagerException: if no import is in progress
        :raise NotFoundException: if branch, repository or import id do not exist
        :raise NotAuthorizedException: if user is not authorized to perform this operation
        :raise ServerException: for any other errors
        """

        if self._import_id is None:
            raise ImportManagerException("No import in progress")

        with api_exception_handler():
            res = self._client.sdk_client.import_api.import_status(repository=self._repo_id,
                                                                   branch=self._branch_id,
                                                                   id=self._import_id)

            if res.completed:
                self._in_progress = False
            return ImportStatus(**res.dict())



import lakefs_sdk

from lakefs import ObjectInfo, CommonPrefix
from lakefs.repository import Repository
from tests.utests.common import get_test_client, expect_exception_context


def get_test_ref():
    client = get_test_client()
    repo = Repository(repository_id="test_repo", client=client)
    return repo.ref("test_reference")


def test_reference_creation():
    ref = get_test_ref()
    assert ref._repo_id == "test_repo"
    assert ref.id == "test_reference"


def test_reference_log(monkeypatch):
    ref = get_test_ref()
    idx = 0
    pages = 10
    items_per_page = 100

    def monkey_log_commits(*_, **__):
        nonlocal idx
        results = []
        for pid in range(items_per_page):
            index = items_per_page * idx + pid
            results.append(lakefs_sdk.Commit(
                id=str(index),
                parents=[""],
                committer="Committer-" + str(index),
                message="Message-" + str(index),
                creation_date=index,
                meta_range_id="",
            ))
        idx += 1
        pagination = lakefs_sdk.Pagination(
            has_more=idx < pages,
            next_offset="",
            max_per_page=items_per_page,
            results=items_per_page
        )

        return lakefs_sdk.CommitList(
            pagination=pagination,
            results=results
        )

    with monkeypatch.context():
        monkeypatch.setattr(ref._client.sdk_client.refs_api, "log_commits", monkey_log_commits)
        i = 0
        # Test log entire history
        for i, c in enumerate(ref.log()):
            assert i == int(c.id)

        assert i + 1 == pages * items_per_page

        # Test log with limit
        idx = 0
        max_amount = 123
        assert len(list(ref.log(max_amount=max_amount))) == max_amount

        # Test limit more than amount
        idx = 0
        max_amount = pages * items_per_page * 2
        assert len(list(ref.log(max_amount=max_amount))) == pages * items_per_page


def test_reference_diff(monkeypatch):
    ref = get_test_ref()
    idx = 0
    pages = 10
    items_per_page = 100

    def monkey_diff_refs(*_, **__):
        nonlocal idx
        results = []
        for pid in range(items_per_page):
            index = items_per_page * idx + pid
            results.append(lakefs_sdk.Diff(
                type="added",
                path=str(index),
                path_type="object",
                size_bytes=index,
            ))
        idx += 1
        pagination = lakefs_sdk.Pagination(
            has_more=idx < pages,
            next_offset="",
            max_per_page=items_per_page,
            results=items_per_page
        )

        return lakefs_sdk.DiffList(
            pagination=pagination,
            results=results
        )

    with monkeypatch.context():
        monkeypatch.setattr(ref._client.sdk_client.refs_api, "diff_refs", monkey_diff_refs)
        # Test log entire history
        i = 0
        for i, c in enumerate(ref.diff(other_ref="other_ref")):
            assert i == c.size_bytes
            assert i == int(c.path)

        assert i + 1 == pages * items_per_page

        # Test log with limit
        idx = 0
        max_amount = 123
        assert len(list(ref.diff(other_ref="other_ref", max_amount=max_amount))) == max_amount

        # Test limit more than amount
        idx = 0
        max_amount = pages * items_per_page * 2
        assert len(list(ref.diff(other_ref="other_ref", max_amount=max_amount))) == pages * items_per_page


def test_reference_objects(monkeypatch):
    ref = get_test_ref()
    with monkeypatch.context():
        def monkey_list_objects(*_, **__):
            results = []
            for i in range(10):
                if i % 2:
                    results.append(lakefs_sdk.ObjectStats(
                        path=f"path-{i}",
                        path_type="object",
                        physical_address=f"address-{i}",
                        checksum=f"{i}",
                        size_bytes=i,
                        mtime=i,
                    ))
                else:
                    results.append(lakefs_sdk.ObjectStats(
                        path=f"path-{i}",
                        path_type="common_prefix",
                        physical_address="?",
                        checksum="",
                        mtime=i,
                    ))
            return lakefs_sdk.ObjectStatsList(pagination=lakefs_sdk.Pagination(
                has_more=False,
                next_offset="",
                max_per_page=1,
                results=1),
                results=results)

        monkeypatch.setattr(ref._client.sdk_client.objects_api, "list_objects", monkey_list_objects)

        for i, item in enumerate(ref.objects()):
            if i % 2:
                assert isinstance(item, ObjectInfo)
                assert item.size_bytes == i
            else:
                assert isinstance(item, CommonPrefix)
                with expect_exception_context(AttributeError):
                    item.checksum  # pylint: disable=pointless-statement

            assert item.path == f"path-{i}"



import ssl
import time
import pytest

from tests.utests.common import expect_exception_context
from lakefs.exceptions import NotFoundException, ConflictException, ObjectNotFoundException
import lakefs


def test_repository_sanity(storage_namespace, setup_repo):
    _, repo = setup_repo
    repo = lakefs.repository(repo.properties.id)  # test the lakefs.repository function works properly
    default_branch = "main"
    expected_properties = lakefs.RepositoryProperties(id=repo.properties.id,
                                                      default_branch=default_branch,
                                                      storage_namespace=storage_namespace,
                                                      creation_date=repo.properties.creation_date)
    assert repo.properties == expected_properties

    # Create with allow exists
    new_data = repo.create(storage_namespace, default_branch, True, exist_ok=True)
    assert new_data.properties == repo.properties

    # Try to create twice and expect conflict
    with expect_exception_context(ConflictException):
        repo.create(storage_namespace, default_branch, True)

    # Get metadata
    _ = repo.metadata

    # List branches
    branches = list(repo.branches())
    assert len(branches) == 1

    # Delete repository
    repo.delete()

    # Delete non existent
    with expect_exception_context(NotFoundException):
        repo.delete()


def test_branch_sanity(setup_repo):
    _, repo = setup_repo
    branch_name = "test_branch"

    main_branch = repo.branch("main")
    new_branch = repo.branch(branch_name).create("main")
    assert new_branch.repo_id == repo.properties.id
    assert new_branch.id == branch_name
    assert new_branch.head.id == main_branch.head.id

    new_branch.delete()
    with expect_exception_context(NotFoundException):
        new_branch.head  # pylint: disable=pointless-statement


def test_ref_sanity(setup_repo):
    _, repo = setup_repo
    ref_id = "main"
    ref = repo.ref(ref_id)
    assert ref.repo_id == repo.properties.id
    assert ref.id == ref_id
    assert ref.get_commit().metadata == {}
    assert ref.get_commit().message == "Repository created"


def test_tag_sanity(setup_repo):
    _, repo = setup_repo
    tag_name = "test_tag"
    tag = repo.tag(tag_name)

    # expect not found
    with expect_exception_context(NotFoundException):
        tag.get_commit()

    branch = repo.branch("main")
    commit = branch.get_commit()
    res = tag.create(commit)
    assert res == tag
    assert tag.id == tag_name
    assert tag.get_commit().metadata == commit.metadata
    assert tag.get_commit().message == commit.message

    # Create again
    with expect_exception_context(ConflictException):
        tag.create(commit.id)

    # Create again with exist_ok
    tag2 = tag.create(tag_name, True)
    assert tag2 == tag

    # Delete tag
    tag.delete()

    # expect not found
    with expect_exception_context(NotFoundException):
        tag.get_commit()

    # Delete twice
    with expect_exception_context(NotFoundException):
        tag.delete()

    # Create again
    tag.create(commit.id)


def test_object_sanity(setup_repo):
    clt, repo = setup_repo
    data = b"test_data"
    path = "test_obj"
    metadata = {"foo": "bar"}
    obj = lakefs.WriteableObject(repository_id=repo.properties.id, reference_id="main", path=path, client=clt).upload(
        data=data, metadata=metadata)
    with obj.reader() as fd:
        assert fd.read() == data

    stats = obj.stat()
    assert stats.path == path == obj.path
    assert stats.mtime <= time.time()
    assert stats.size_bytes == len(data)
    assert stats.metadata == metadata
    assert stats.content_type == "application/octet-stream"

    obj.delete()
    with expect_exception_context(ObjectNotFoundException):
        obj.stat()


@pytest.mark.parametrize("ssl_enabled", (True, False, None))
def test_ssl_configuration(ssl_enabled):
    expected_ssl = ssl_enabled is not False
    expected_cert_req = ssl.CERT_REQUIRED if expected_ssl else ssl.CERT_NONE
    proxy = "http://someproxy:1234"
    cert = "somecert"
    clt = lakefs.client.Client(verify_ssl=ssl_enabled, proxy=proxy, ssl_ca_cert=cert)
    assert clt.config.verify_ssl is expected_ssl
    assert clt.config.proxy == proxy
    assert clt.config.ssl_ca_cert == cert
    assert clt.sdk_client._api.rest_client.pool_manager.connection_pool_kw.get("cert_reqs") is expected_cert_req
    assert str(clt.sdk_client._api.rest_client.pool_manager.proxy) == proxy
    assert clt.sdk_client._api.rest_client.pool_manager.connection_pool_kw.get("ca_certs") is cert



"""
Module containing all of lakeFS data models
"""

from __future__ import annotations

from datetime import datetime
from typing import List, Optional, Literal

from lakefs.namedtuple import LenientNamedTuple

_COMMON_PREFIX = "common_prefix"
_OBJECT = "object"


class Commit(LenientNamedTuple):
    """
    NamedTuple representing a lakeFS commit's properties
    """
    id: str
    parents: List[str]
    committer: str
    message: str
    creation_date: int
    meta_range_id: str
    metadata: Optional[dict[str, str]] = None


class Change(LenientNamedTuple):
    """
    NamedTuple representing a diff change between two refs in lakeFS
    """
    type: Literal["added", "removed", "changed", "conflict", "prefix_changed"]
    path: str
    path_type: Literal["common_prefix", "object"]
    size_bytes: Optional[int]

    def __repr__(self):
        return f'Change(type="{self.type}", path="{self.path}", path_type="{self.path_type}")'


class ImportStatus(LenientNamedTuple):
    """
    NamedTuple representing an ongoing import's status in lakeFS
    """

    class _Error(LenientNamedTuple):
        message: str

    completed: bool
    update_time: datetime
    ingested_objects: Optional[int]
    metarange_id: Optional[str]
    commit: Optional[Commit]
    error: Optional[_Error]

    def __init__(self, **kwargs):
        commit = kwargs.get("commit")
        if commit is not None:
            kwargs["commit"] = Commit(**commit)

        error = kwargs.get("error")
        if error is not None:
            kwargs["error"] = ImportStatus._Error(**error)

        super().__init__(**kwargs)


class ServerStorageConfiguration(LenientNamedTuple):
    """
    Represent a lakeFS server's storage configuration
    """
    blockstore_type: str
    pre_sign_support: bool
    import_support: bool
    blockstore_namespace_example: str
    blockstore_namespace_validity_regex: str
    pre_sign_support_ui: bool
    import_validity_regex: str
    default_namespace_prefix: Optional[str] = None


class ObjectInfo(LenientNamedTuple):
    """
    Represent a lakeFS object's stats
    """
    path: str
    physical_address: str
    checksum: str
    mtime: int
    physical_address_expiry: Optional[int] = None
    size_bytes: Optional[int] = None
    metadata: Optional[dict[str, str]] = None
    content_type: Optional[str] = None

    def __repr__(self):
        return f'ObjectInfo(path="{self.path}")'


class CommonPrefix(LenientNamedTuple):
    """
    Represents a common prefix in lakeFS
    """
    path: str

    def __repr__(self):
        return f'CommonPrefix(path="{self.path}")'


class RepositoryProperties(LenientNamedTuple):
    """
    Represent a lakeFS repository's properties
    """
    id: str
    creation_date: int
    default_branch: str
    storage_namespace: str



import uuid
import pytest

import lakefs

from tests.integration.conftest import _setup_repo, get_storage_namespace

_NUM_PREFIXES = 10
_NUM_ELEM_PER_PREFIX = 20


@pytest.fixture(name="setup_repo_with_branches_and_tags", scope="session")
def fixture_setup_repo_with_branches_and_tags():
    clt, repo = _setup_repo(get_storage_namespace("branches-and-tags"),
                            "branches-and-tags",
                            "main")
    for i in range(_NUM_PREFIXES):
        for j in range(_NUM_ELEM_PER_PREFIX):
            b = repo.branch(f"branches{i:02d}-{j:02d}").create("main")
            repo.tag(tag_id=f"tags{i:02d}-{j:02d}").create(b)

    return clt, repo


@pytest.mark.parametrize("attr", ("branches", "tags"))
def test_repository_listings(setup_repo_with_branches_and_tags, attr):
    _, repo = setup_repo_with_branches_and_tags

    generator = getattr(repo, attr)

    total = _NUM_PREFIXES * _NUM_ELEM_PER_PREFIX
    if attr == "branches":
        total += 1  # Including main
    assert len(list(generator())) == total

    after = 9
    res = list(generator(max_amount=100, prefix=f"{attr}01", after=f"{attr}01-{after:02d}"))
    assert len(res) == 10
    for i, b in enumerate(res):
        assert b.id == f"{attr}01-{i + after + 1:02d}"


def test_repositories(storage_namespace):
    repo_base_name = f"test-repo{uuid.uuid4()}-"
    for i in range(10):
        lakefs.repository(f"{repo_base_name}{i}").create(storage_namespace=f"{storage_namespace}-{i}")

    repos = list(lakefs.repositories(prefix=repo_base_name))
    assert len(repos) == 10
    for i, repo in enumerate(repos):
        assert repo.properties.id == f"{repo_base_name}{i}"



import http
import os
import time

import lakefs_sdk

from tests.utests.common import (
    get_test_repo,
    expect_exception_context,
    lakectl_no_config_context,
    env_var_context,
    TEST_REPO_ARGS
)

from lakefs.exceptions import (
    ServerException,
    NotAuthorizedException,
    NotFoundException,
    ConflictException,
    NoAuthenticationFound
)
from lakefs import RepositoryProperties


def monkey_create_repository(_self, repository_creation, *_):
    assert repository_creation.name == TEST_REPO_ARGS.name
    assert repository_creation.storage_namespace == TEST_REPO_ARGS.storage_namespace
    assert repository_creation.default_branch == TEST_REPO_ARGS.default_branch
    assert repository_creation.sample_data == TEST_REPO_ARGS.sample_data
    return lakefs_sdk.Repository(id=TEST_REPO_ARGS.name,
                                 creation_date=int(time.time()),
                                 storage_namespace=TEST_REPO_ARGS.storage_namespace,
                                 default_branch=TEST_REPO_ARGS.default_branch)


def test_repository_creation(monkeypatch):
    repo = get_test_repo()

    with monkeypatch.context():
        monkeypatch.setattr(lakefs_sdk.RepositoriesApi, "create_repository", monkey_create_repository)
        repo.create(storage_namespace=TEST_REPO_ARGS.storage_namespace,
                    default_branch=TEST_REPO_ARGS.default_branch,
                    include_samples=TEST_REPO_ARGS.sample_data)


def test_repository_creation_already_exists(monkeypatch):
    repo = get_test_repo()
    ex = lakefs_sdk.exceptions.ApiException(status=http.HTTPStatus.CONFLICT.value)

    with monkeypatch.context():
        def monkey_create_repository_ex(*_):
            raise ex

        monkeypatch.setattr(lakefs_sdk.RepositoriesApi, "create_repository", monkey_create_repository_ex)

        # Expect success when exist_ok = True
        existing = lakefs_sdk.Repository(id=TEST_REPO_ARGS.name,
                                         default_branch="main",
                                         storage_namespace="s3://existing-namespace",
                                         creation_date=12345)

        def monkey_get_repository(*_):
            return existing

        monkeypatch.setattr(lakefs_sdk.RepositoriesApi, "get_repository", monkey_get_repository)
        res = repo.create(storage_namespace=TEST_REPO_ARGS.storage_namespace,
                          default_branch=TEST_REPO_ARGS.default_branch,
                          include_samples=TEST_REPO_ARGS.sample_data,
                          exist_ok=True)

        assert res.properties == RepositoryProperties(**existing.dict())

    # Expect fail on exists
    with expect_exception_context(ConflictException):
        repo.create(storage_namespace=TEST_REPO_ARGS.storage_namespace,
                    default_branch=TEST_REPO_ARGS.default_branch,
                    include_samples=TEST_REPO_ARGS.sample_data)

    # Expect fail on exists
    ex = lakefs_sdk.exceptions.UnauthorizedException(http.HTTPStatus.UNAUTHORIZED)
    with expect_exception_context(NotAuthorizedException):
        repo.create(storage_namespace=TEST_REPO_ARGS.storage_namespace,
                    default_branch=TEST_REPO_ARGS.default_branch,
                    include_samples=TEST_REPO_ARGS.sample_data)


def test_delete_repository(monkeypatch):
    repo = get_test_repo()
    with monkeypatch.context():
        monkeypatch.setattr(lakefs_sdk.RepositoriesApi, "delete_repository", lambda *args: None)
        repo.delete()

        ex = None

        def monkey_delete_repository(*_):
            raise ex

        monkeypatch.setattr(lakefs_sdk.RepositoriesApi, "delete_repository", monkey_delete_repository)
        # Not found
        ex = lakefs_sdk.exceptions.NotFoundException(status=http.HTTPStatus.NOT_FOUND)
        with expect_exception_context(NotFoundException):
            repo.delete()

        # Unauthorized
        ex = lakefs_sdk.exceptions.UnauthorizedException(status=http.HTTPStatus.UNAUTHORIZED)
        with expect_exception_context(NotAuthorizedException):
            repo.delete()

        # Other error
        ex = lakefs_sdk.exceptions.ApiException()
        with expect_exception_context(ServerException):
            repo.delete()


def test_create_repository_no_authentication(monkeypatch):
    with lakectl_no_config_context(monkeypatch):
        from lakefs.repository import Repository
        repo = Repository("test-repo", None)
        with expect_exception_context(NoAuthenticationFound):
            repo.create(storage_namespace=TEST_REPO_ARGS.storage_namespace)

        # update credentials and retry create
        monkeypatch.setattr(lakefs_sdk.RepositoriesApi, "create_repository", monkey_create_repository)
        with env_var_context():
            from lakefs import config as client_config
            # Create a new client with env vars
            os.environ[client_config._LAKECTL_ENDPOINT_ENV] = "endpoint"
            os.environ[client_config._LAKECTL_SECRET_ACCESS_KEY_ENV] = "secret"
            os.environ[client_config._LAKECTL_ACCESS_KEY_ID_ENV] = "key"
            repo.create(storage_namespace=TEST_REPO_ARGS.storage_namespace,
                        default_branch=TEST_REPO_ARGS.default_branch)



"""
lakeFS Client module

Handles authentication against the lakeFS server and wraps the underlying lakefs_sdk client.
"""

from __future__ import annotations

import base64
import json
from threading import Lock
from typing import Optional
from typing import TYPE_CHECKING
from urllib.parse import urlparse, parse_qs

import lakefs_sdk
from lakefs_sdk import ExternalLoginInformation
from lakefs_sdk.client import LakeFSClient

from lakefs.config import ClientConfig
from lakefs.exceptions import NotAuthorizedException, ServerException, api_exception_handler
from lakefs.models import ServerStorageConfiguration

if TYPE_CHECKING:
    import boto3

DEFAULT_REGION = 'us-east-1'


class ServerConfiguration:
    """
    Represent a lakeFS server's configuration
    """
    _conf: lakefs_sdk.Config
    _storage_conf: ServerStorageConfiguration

    def __init__(self, client: Optional[Client] = None):
        try:
            self._conf = client.sdk_client.config_api.get_config()
            self._storage_conf = ServerStorageConfiguration(**self._conf.storage_config.dict())
        except lakefs_sdk.exceptions.ApiException as e:
            if isinstance(e, lakefs_sdk.exceptions.ApiException):
                raise NotAuthorizedException(e.status, e.reason) from e
            raise ServerException(e.status, e.reason) from e

    @property
    def version(self) -> str:
        """
        Return the lakeFS server version
        """
        return self._conf.version_config.version

    @property
    def storage_config(self) -> ServerStorageConfiguration:
        """
        Returns the lakeFS server storage configuration
        """
        return self._storage_conf


class Client:
    """
    Wrapper around lakefs_sdk's client object
    Takes care of instantiating it from the environment

    Example of initializing a Client object:

    .. code-block:: python

        from lakefs import Client

        client = Client(username="<access_key_id>", password="<secret_access_key>", host="<lakefs_endpoint>")
        print(client.version)

    """

    _client: Optional[LakeFSClient] = None
    _conf: Optional[ClientConfig] = None
    _server_conf: Optional[ServerConfiguration] = None

    def __init__(self, **kwargs):
        self._conf = ClientConfig(**kwargs)
        self._client = LakeFSClient(self._conf, header_name='X-Lakefs-Client',
                                    header_value='python-lakefs')

    @property
    def config(self):
        """
        Return the underlying lakefs_sdk configuration
        """
        return self._conf

    @property
    def sdk_client(self):
        """
        Return the underlying lakefs_sdk client
        """
        return self._client

    @property
    def storage_config(self):
        """
        lakeFS SDK storage config object, lazy evaluated.
        """
        if self._server_conf is None:
            self._server_conf = ServerConfiguration(self)
        return self._server_conf.storage_config

    @property
    def version(self) -> str:
        """
        lakeFS Server version, lazy evaluated.
        """
        if self._server_conf is None:
            self._server_conf = ServerConfiguration(self)
        return self._server_conf.version


def _extract_region_from_endpoint(endpoint):
    """
    Extract the region name from an STS endpoint URL.
    for example: https://sts.eu-central-1.amazonaws.com/ -> eu-central-1
    and for example: https://sts.amazonaws.com/ -> DEFAULT_REGION

    :param endpoint: The endpoint URL of the STS client.
    :return: The region name extracted from the endpoint URL.
    """

    parts = endpoint.split('.')
    if len(parts) == 4:
        return parts[1]
    if len(parts) > 4:
        return parts[2]
    return DEFAULT_REGION


def _get_identity_token(
        session: boto3.Session,
        lakefs_host: str,
        additional_headers: dict[str, str],
        presign_expiry
) -> str:
    """
   Generate the identity token required for lakeFS authentication from an AWS session.

   This function uses the STS client to generate a presigned URL for the `get_caller_identity` action,
    extracts the required values from the URL,
   and creates a base64-encoded JSON object with these values.

   :param session: A boto3 session object with the necessary AWS credentials and region information.
   :return: A base64-encoded JSON string containing the required authentication information.
   :raises ValueError: If the session does not have a region name set.
   """

    # this method should only be called when installing the aws-iam additional requirement
    from botocore.client import Config  # pylint: disable=import-outside-toplevel, import-error
    from botocore.signers import RequestSigner  # pylint: disable=import-outside-toplevel, import-error

    sts_client = session.client('sts', config=Config(signature_version='v4'))
    endpoint = sts_client.meta.endpoint_url
    service_id = sts_client.meta.service_model.service_id
    region = _extract_region_from_endpoint(endpoint)
    # signer is used because the presigned URL generated by the STS does not support additional headers
    signer = RequestSigner(
        service_id,
        region,
        'sts',
        'v4',
        session.get_credentials(),
        session.events
    )
    endpoint_with_params = f"{endpoint}/?Action=GetCallerIdentity&Version=2011-06-15"
    if additional_headers is None:
        additional_headers = {
            'X-LakeFS-Server-ID': lakefs_host,
        }
    params = {
        'method': 'POST',
        'url': endpoint_with_params,
        'body': {},
        'headers': additional_headers,
        'context': {}
    }

    presigned_url = signer.generate_presigned_url(
        params,
        region_name=region,
        expires_in=presign_expiry,
        operation_name=''
    )
    parsed_url = urlparse(presigned_url)
    query_params = parse_qs(parsed_url.query)

    # Extract values from query parameters
    json_object = {
        "method": "POST",
        "host": parsed_url.hostname,
        "region": region,
        "action": query_params['Action'][0],
        "date": query_params['X-Amz-Date'][0],
        "expiration_duration": query_params['X-Amz-Expires'][0],
        "access_key_id": query_params['X-Amz-Credential'][0].split('/')[0],
        "signature": query_params['X-Amz-Signature'][0],
        "signed_headers": query_params.get('X-Amz-SignedHeaders', [''])[0].split(';'),
        "version": query_params['Version'][0],
        "algorithm": query_params['X-Amz-Algorithm'][0],
        "security_token": query_params.get('X-Amz-Security-Token', [None])[0]
    }

    json_string = json.dumps(json_object)
    return base64.b64encode(json_string.encode('utf-8')).decode('utf-8')


def from_aws_role(
        session: boto3.Session,
        ttl_seconds: int = 3600,
        presigned_ttl: int = 60,
        additional_headers: dict[str, str] = None,
        **kwargs) -> Client:
    """
    Create a lakeFS client from an AWS role.
    :param session: : The boto3 session.
    :param ttl_seconds: The time-to-live for the generated lakeFS token in seconds. The default value is 3600 seconds.
    :param presigned_ttl: The time-to-live for the presigned URL in seconds. The default value is 60 seconds.
    :param additional_headers: Additional headers to include in the presigned URL.
    :param kwargs: The arguments to pass to the client.
    :return: A lakeFS client.
    """

    client = Client(**kwargs)
    lakefs_host = urlparse(client.config.host).hostname
    identity_token = _get_identity_token(session, lakefs_host, presign_expiry=presigned_ttl,
                                         additional_headers=additional_headers)
    external_login_information = ExternalLoginInformation(token_expiration_duration=ttl_seconds, identity_request={
        "identity_token": identity_token
    })

    with api_exception_handler():
        auth_token = client.sdk_client.auth_api.external_principal_login(external_login_information)

    client.config.access_token = auth_token.token
    return client


def from_web_identity(code: str, state: str, redirect_uri: str, ttl_seconds: int = 3600, **kwargs) -> Client:
    """
    Authenticate against lakeFS using a code received from an identity provider

    :param code: The code received from the identity provider
    :param state: The state received from the identity provider
    :param redirect_uri: The redirect URI used in the authentication process
    :param ttl_seconds: The token's time-to-live in seconds
    :param kwargs: Remaining arguments for the Client object
    :return: The authenticated Client object
    :raise NotAuthorizedException: if user is not authorized to perform this operation
    """
    client = Client(**kwargs)
    sts_requests = lakefs_sdk.StsAuthRequest(code=code, state=state, redirect_uri=redirect_uri, ttl_seconds=ttl_seconds)
    with api_exception_handler():
        auth_token = client.sdk_client.experimental_api.sts_login(sts_requests)
    client.config.access_token = auth_token.token
    return client


class _BaseLakeFSObject:
    """
    Base class for all lakeFS SDK objects, holds the client object and handles errors where no authentication method
    found for client. Attempts to reload client dynamically in case of changes in the environment.
    """
    __mutex: Lock = Lock()
    __client: Optional[Client] = None

    def __init__(self, client: Optional[Client]):
        self.__client = client

    @property
    def _client(self):
        """
        If client is None due to missing authentication params, try to init again. If authentication method is still
        missing - will raise exception
        :return: The initialized client object
        :raise NoAuthenticationFound: If no authentication method found to configure the lakeFS client with
        """
        if self.__client is not None:
            return self.__client

        with _BaseLakeFSObject.__mutex:
            if _BaseLakeFSObject.__client is None:
                _BaseLakeFSObject.__client = Client()
            return _BaseLakeFSObject.__client



"""
Module containing lakeFS branch implementation
"""

from __future__ import annotations

import uuid
import warnings
from contextlib import contextmanager
from typing import Optional, Generator, Iterable, Literal, Dict

import lakefs_sdk
from lakefs.client import Client
from lakefs.object import WriteableObject
from lakefs.object import StoredObject
from lakefs.import_manager import ImportManager
from lakefs.reference import Reference, ReferenceType, generate_listing
from lakefs.models import Change, Commit
from lakefs.exceptions import (
    api_exception_handler,
    ConflictException,
    LakeFSException,
    TransactionException
)


class _BaseBranch(Reference):

    def object(self, path: str) -> WriteableObject:
        """
        Returns a writable object using the current repo id, reference and path

        :param path: The object's path
        """

        return WriteableObject(self.repo_id, self._id, path, client=self._client)

    def uncommitted(self, max_amount: Optional[int] = None, after: Optional[str] = None, prefix: Optional[str] = None,
                    **kwargs) -> Generator[Change]:
        """
        Returns a diff generator of uncommitted changes on this branch

        :param max_amount: Stop showing changes after this amount
        :param after: Return items after this value
        :param prefix: Return items prefixed with this value
        :param kwargs: Additional Keyword Arguments to send to the server
        :raise NotFoundException: if branch or repository do not exist
        :raise NotAuthorizedException: if user is not authorized to perform this operation
        :raise ServerException: for any other errors
        """

        for diff in generate_listing(self._client.sdk_client.branches_api.diff_branch,
                                     self._repo_id, self._id, max_amount=max_amount, after=after, prefix=prefix,
                                     **kwargs):
            yield Change(**diff.dict())

    def delete_objects(self, object_paths: str | StoredObject | Iterable[str | StoredObject]) -> None:
        """
        Delete objects from lakeFS

        This method can be used to delete single/multiple objects from branch. It accepts both str and StoredObject
        types as well as Iterables of these types.
        Using this method is more performant than sequentially calling delete on objects as it saves the back and forth
        from the server.

        This can also be used in combination with object listing. For example:

        .. code-block:: python

            import lakefs

            branch = lakefs.repository("<repository_name>").branch("<branch_name>")
            # list objects on a common prefix
            objs = branch.objects(prefix="my-object-prefix/", max_amount=100)
            # delete objects which have "foo" in their name
            branch.delete_objects([o.path for o in objs if "foo" in o.path])

        :param object_paths: a single path or an iterable of paths to delete
        :raise NotFoundException: if branch or repository do not exist
        :raise NotAuthorizedException: if user is not authorized to perform this operation
        :raise ServerException: for any other errors
        """
        if isinstance(object_paths, str):
            object_paths = [object_paths]
        elif isinstance(object_paths, StoredObject):
            object_paths = [object_paths.path]
        elif isinstance(object_paths, Iterable):
            object_paths = [o.path if isinstance(o, StoredObject) else o for o in object_paths]
        with api_exception_handler():
            return self._client.sdk_client.objects_api.delete_objects(
                self._repo_id,
                self._id,
                lakefs_sdk.PathList(paths=object_paths)
            )

    def reset_changes(self, path_type: Literal["common_prefix", "object", "reset"] = "reset",
                      path: Optional[str] = None) -> None:
        """
        Reset uncommitted changes (if any) on this branch

        :param path_type: the type of path to reset ('common_prefix', 'object', 'reset' - for all changes)
        :param path: the path to reset (optional) - if path_type is 'reset' this parameter is ignored
        :raise ValidationError: if path_type is not one of the allowed values
        :raise NotFoundException: if branch or repository do not exist
        :raise NotAuthorizedException: if user is not authorized to perform this operation
        :raise ServerException: for any other errors
        """

        reset_creation = lakefs_sdk.ResetCreation(path=path, type=path_type)
        return self._client.sdk_client.branches_api.reset_branch(self._repo_id, self.id, reset_creation)


class Branch(_BaseBranch):
    """
    Class representing a branch in lakeFS.
    """

    def __init__(self, repository_id: str, branch_id: str, client: Optional[Client] = None):
        super().__init__(repository_id, reference_id=branch_id, client=client)

    def get_commit(self):
        """
        For branches override the default _get_commit method to ensure we always fetch the latest head
        """
        self._commit = None
        return super().get_commit()

    def cherry_pick(self, reference: ReferenceType, parent_number: Optional[int] = None) -> Commit:
        """
        Cherry-pick a given reference onto the branch.

        :param reference: ID of the reference to cherry-pick.
        :param parent_number: When cherry-picking a merge commit, the parent number (starting from 1)
            with which to perform the diff. The default branch is parent 1.
        :return: The cherry-picked commit at the head of the branch.
        :raise NotFoundException: If either the repository or target reference do not exist.
        :raise NotAuthorizedException: If the user is not authorized to perform this operation.
        :raise ServerException: For any other errors.
        """
        ref = reference if isinstance(reference, str) else reference.id
        cherry_pick_creation = lakefs_sdk.CherryPickCreation(ref=ref, parent_number=parent_number)
        with api_exception_handler():
            res = self._client.sdk_client.branches_api.cherry_pick(self._repo_id, self._id, cherry_pick_creation)
            return Commit(**res.dict())

    def create(self, source_reference: ReferenceType, exist_ok: bool = False) -> Branch:
        """
        Create a new branch in lakeFS from this object

        Example of creating a new branch:

        .. code-block:: python

            import lakefs

            branch = lakefs.repository("<repository_name>").branch("<branch_name>").create("<source_reference>")

        :param source_reference: The reference to create the branch from (reference ID, object or Commit object)
        :param exist_ok: If False will throw an exception if a branch by this name already exists. Otherwise,
            return the existing branch without creating a new one
        :return: The lakeFS SDK object representing the branch
        :raise NotFoundException: if repo, branch or source reference id does not exist
        :raise ConflictException: if branch already exists and exist_ok is False
        :raise NotAuthorizedException: if user is not authorized to perform this operation
        :raise ServerException: for any other errors
        """

        def handle_conflict(e: LakeFSException):
            if isinstance(e, ConflictException) and exist_ok:
                return None
            return e

        reference_id = source_reference if isinstance(source_reference, str) else source_reference.id
        branch_creation = lakefs_sdk.BranchCreation(name=self._id, source=reference_id)
        with api_exception_handler(handle_conflict):
            self._client.sdk_client.branches_api.create_branch(self._repo_id, branch_creation)
        return self

    @property
    def head(self) -> Reference:
        """
        Get the commit reference this branch is pointing to

        :return: The commit reference this branch is pointing to
        :raise NotFoundException: if branch by this id does not exist
        :raise NotAuthorizedException: if user is not authorized to perform this operation
        :raise ServerException: for any other errors
        """
        with api_exception_handler():
            branch = self._client.sdk_client.branches_api.get_branch(self._repo_id, self._id)
        return Reference(self._repo_id, branch.commit_id, self._client)

    def commit(self, message: str, metadata: dict = None, **kwargs) -> Reference:
        """
        Commit changes on the current branch

        :param message: Commit message
        :param metadata: Metadata to attach to the commit
        :param kwargs: Additional Keyword Arguments for commit creation
        :return: The new reference after the commit
        :raise NotFoundException: if branch by this id does not exist
        :raise ForbiddenException: if commit is not allowed on this branch
        :raise NotAuthorizedException: if user is not authorized to perform this operation
        :raise ServerException: for any other errors
        """
        commits_creation = lakefs_sdk.CommitCreation(message=message, metadata=metadata, **kwargs)

        with api_exception_handler():
            c = self._client.sdk_client.commits_api.commit(self._repo_id, self._id, commits_creation)
        return Reference(self._repo_id, c.id, self._client)

    def delete(self) -> None:
        """
        Delete branch from lakeFS server

        :raise NotFoundException: if branch or repository do not exist
        :raise NotAuthorizedException: if user is not authorized to perform this operation
        :raise ForbiddenException: for branches that are protected
        :raise ServerException: for any other errors
        """
        with api_exception_handler():
            return self._client.sdk_client.branches_api.delete_branch(self._repo_id, self._id)

    def revert(self, reference: Optional[ReferenceType], parent_number: int = 0, *,
               reference_id: Optional[str] = None) -> Commit:
        """
        revert the changes done by the provided reference on the current branch

        :param reference_id: (Optional) The reference ID to revert

            .. deprecated:: 0.4.0
                Use ``reference`` instead.

        :param parent_number: when reverting a merge commit, the parent number (starting from 1) relative to which to
            perform the revert. The default for non merge commits is 0
        :param reference: the reference to revert
        :return: The commit created by the revert
        :raise NotFoundException: if branch by this id does not exist
        :raise NotAuthorizedException: if user is not authorized to perform this operation
        :raise ServerException: for any other errors
        """
        if parent_number < 0:
            raise ValueError("parent_number must be a non-negative integer")

        if reference_id is not None:
            warnings.warn(
                "reference_id is deprecated, please use the `reference` argument.", DeprecationWarning
            )

        # Handle reference_id as a deprecated alias to reference.
        reference = reference or reference_id
        if reference is None:
            raise ValueError("reference to revert must be specified")
        ref = reference if isinstance(reference, str) else reference.id

        with api_exception_handler():
            self._client.sdk_client.branches_api.revert_branch(
                self._repo_id,
                self._id,
                lakefs_sdk.RevertCreation(ref=ref, parent_number=parent_number)
            )
            commit = self._client.sdk_client.commits_api.get_commit(self._repo_id, self._id)
            return Commit(**commit.dict())

    def import_data(self, commit_message: str = "", metadata: Optional[dict] = None) -> ImportManager:
        """
        Import data to lakeFS

        :param metadata: metadata to attach to the commit
        :param commit_message: once the data is imported, a commit is created with this message. If default (empty)
            message is provided, uses the default server commit message for imports.
        :return: an ImportManager object
        """
        return ImportManager(self._repo_id, self._id, commit_message, metadata, self._client)

    @contextmanager
    def transact(self, commit_message: str = "", commit_metadata: Optional[Dict] = None,
                 delete_branch_on_error: bool = True) -> _Transaction:
        """
        Create a transaction for multiple operations.
        Transaction allows for multiple modifications to be performed atomically on a branch,
        similar to a database transaction.
        It ensures that the branch remains unaffected until the transaction is successfully completed.
        The process includes:

        1. Creating an ephemeral branch from this branch
        2. Perform object operations on ephemeral branch
        3. Commit changes
        4. Merge back to source branch
        5. Delete ephemeral branch

        Using a transaction the code for this flow will look like this:

        .. code-block:: python

            import lakefs

            branch = lakefs.repository("<repository_name>").branch("<branch_name>")
            with branch.transact(commit_message="my transaction") as tx:
                for obj in tx.objects(prefix="prefix_to_delete/"):  # Delete some objects
                    obj.delete()

                # Create new object
                tx.object("new_object").upload("new object data")

        Note that unlike database transactions, lakeFS transaction does not take a "lock" on the branch, and therefore
        the transaction might fail due to changes in source branch after the transaction was created.

        :param commit_message: once the transaction is committed, a commit is created with this message
        :param commit_metadata: user metadata for the transaction commit
        :param delete_branch_on_error: Defaults to True. Ensures ephemeral branch is deleted on error.
        :return: a Transaction object to perform the operations on
        """
        with Transaction(self._repo_id, self._id, commit_message, commit_metadata, delete_branch_on_error,
                         self._client) as tx:
            yield tx


class _Transaction(_BaseBranch):
    @staticmethod
    def _get_tx_name() -> str:
        return f"tx-{uuid.uuid4()}"  # Don't rely on source branch name as this might exceed valid branch length

    def __init__(self, repository_id: str, branch_id: str, commit_message: str = "",
                 commit_metadata: Optional[Dict] = None, client: Client = None):
        self._commit_message = commit_message
        self._commit_metadata = commit_metadata
        self._source_branch = branch_id

        tx_name = self._get_tx_name()
        self._tx_branch = Branch(repository_id, tx_name, client).create(branch_id)
        super().__init__(repository_id, tx_name, client)

    @property
    def source_id(self) -> str:
        """
        Returns the source branch ID the transaction is associated to
        """
        return self._source_branch

    @property
    def commit_message(self) -> str:
        """
        Return the commit message configured for this transaction completion
        """
        return self._commit_message

    @commit_message.setter
    def commit_message(self, message: str) -> None:
        """
        Set the commit message for this transaction completion
        :param message: The commit message to use on the transaction merge commit
        """
        self._commit_message = message

    @property
    def commit_metadata(self) -> Optional[Dict]:
        """
        Return the commit metadata configured for this transaction completion
        """
        return self._commit_metadata

    @commit_metadata.setter
    def commit_metadata(self, metadata: Optional[Dict]) -> None:
        """
        Set the commit metadata for this transaction completion
        :param metadata: The metadata to use on the transaction merge commit
        """
        self._commit_metadata = metadata


class Transaction:
    """
    Manage a transaction on a given branch

    The transaction creates an ephemeral branch from the source branch. The transaction can then be used to perform
    operations on the branch which will later be merged back into the source branch.
    Currently, transaction is supported only as a context manager.
    """

    def __init__(self, repository_id: str, branch_id: str, commit_message: str = "",
                 commit_metadata: Optional[Dict] = None, delete_branch_on_error: bool = True, client: Client = None):
        self._repo_id = repository_id
        self._commit_message = commit_message
        self._commit_metadata = commit_metadata
        self._source_branch = branch_id
        self._client = client
        self._tx = None
        self._tx_branch = None
        self._cleanup_branch = delete_branch_on_error

    def __enter__(self):
        self._tx = _Transaction(self._repo_id, self._source_branch, self._commit_message, self._commit_metadata,
                                self._client)
        self._tx_branch = Branch(self._repo_id, self._tx.id, self._client)
        return self._tx

    def __exit__(self, typ, value, traceback) -> bool:
        if typ is not None:  # Perform only cleanup in case exception occurred
            if self._cleanup_branch:
                self._tx_branch.delete()
            return False  # Raise the underlying exception

        try:
            self._tx_branch.commit(message=self._tx.commit_message, metadata=self._tx.commit_metadata)
            self._tx_branch.merge_into(self._source_branch, message=f"Merge transaction {self._tx.id} to branch")
            self._tx_branch.delete()

            return False
        except LakeFSException as e:
            if self._cleanup_branch:
                self._tx_branch.delete()
            raise TransactionException(f"Failed committing transaction {self._tx.id}: {e}") from e



"""
Client configuration module
"""

from __future__ import annotations

import os
from pathlib import Path
from typing import Optional

import yaml
from lakefs_sdk import Configuration
from lakefs.exceptions import NoAuthenticationFound
from lakefs.namedtuple import LenientNamedTuple

_LAKECTL_YAML_PATH = os.path.join(Path.home(), ".lakectl.yaml")
_LAKECTL_ENDPOINT_ENV = "LAKECTL_SERVER_ENDPOINT_URL"
_LAKECTL_ACCESS_KEY_ID_ENV = "LAKECTL_CREDENTIALS_ACCESS_KEY_ID"
_LAKECTL_SECRET_ACCESS_KEY_ENV = "LAKECTL_CREDENTIALS_SECRET_ACCESS_KEY"


class ClientConfig(Configuration):
    """
    Configuration class for the SDK Client.
    Instantiation will try to get authentication methods using the following chain:

    1. Provided kwargs to __init__ func (should contain necessary credentials as defined in lakefs_sdk.Configuration)
    2. Use LAKECTL_SERVER_ENDPOINT_URL, LAKECTL_ACCESS_KEY_ID and LAKECTL_ACCESS_SECRET_KEY if set
    3. Try to read ~/.lakectl.yaml if exists
    4. TBD: try and use IAM role from current machine (using AWS IAM role will work only with enterprise/cloud)

    This class also encapsulates the required lakectl configuration for authentication and used to unmarshall the
    lakectl yaml file.
    """

    class Server(LenientNamedTuple):
        """
        lakectl configuration's server block
        """
        endpoint_url: str

    class Credentials(LenientNamedTuple):
        """
        lakectl configuration's credentials block
        """
        access_key_id: str
        secret_access_key: str

    server: Server
    credentials: Credentials

    def __init__(self, verify_ssl: Optional[bool] = None, proxy: Optional[str] = None, **kwargs):
        super().__init__(**kwargs)
        if verify_ssl is not None:
            self.verify_ssl = verify_ssl
        if proxy is not None:
            self.proxy = proxy

        if kwargs:
            return

        found = False

        # Get credentials from lakectl
        try:
            with open(_LAKECTL_YAML_PATH, encoding="utf-8") as fd:
                data = yaml.load(fd, Loader=yaml.Loader)
                self.server = ClientConfig.Server(**data["server"])
                self.credentials = ClientConfig.Credentials(**data["credentials"])
            found = True
        except FileNotFoundError:  # File not found, fallback to env variables
            self.server = ClientConfig.Server(endpoint_url="")
            self.credentials = ClientConfig.Credentials(access_key_id="", secret_access_key="")

        endpoint_env = os.getenv(_LAKECTL_ENDPOINT_ENV)
        key_env = os.getenv(_LAKECTL_ACCESS_KEY_ID_ENV)
        secret_env = os.getenv(_LAKECTL_SECRET_ACCESS_KEY_ENV)

        self.host = endpoint_env if endpoint_env is not None else self.server.endpoint_url
        self.username = key_env if key_env is not None else self.credentials.access_key_id
        self.password = secret_env if secret_env is not None else self.credentials.secret_access_key
        if len(self.username) > 0 and len(self.password) > 0:
            found = True

        # TODO: authentication via IAM Role
        if not found:
            raise NoAuthenticationFound



from tests.utests.common import get_test_client

from lakefs.repository import Repository


def test_tag_creation():
    """
    Ensure tags can only be created in repo context
    """
    client = get_test_client()
    repo = Repository(repository_id="test_repo", client=client)
    tag = repo.tag("test_tag")
    assert tag.repo_id == "test_repo"
    assert tag.id == "test_tag"



"""
Allow importing of models from package root
"""

from lakefs.client import Client
from lakefs.repository import Repository, repositories
from lakefs.reference import Reference
from lakefs.models import (
    Commit,
    Change,
    ImportStatus,
    ServerStorageConfiguration,
    ObjectInfo,
    CommonPrefix,
    RepositoryProperties
)
from lakefs.tag import Tag
from lakefs.branch import Branch
from lakefs.object import StoredObject, WriteableObject, ObjectReader


def repository(repository_id: str, *args, **kwargs) -> Repository:
    """
    Wrapper for getting a Repository object from the lakefs module.
    Enable more fluid syntax (lakefs.repository("x").branch("y") instead of lakefs.Repository("x").branch("y"))

    :param repository_id: The repository name
    :return: Repository object representing a lakeFS repository with the give repository_id
    """
    return Repository(repository_id, *args, **kwargs)



from setuptools import setup, find_packages

NAME = "lakefs"
VERSION = "0.7.0"
# To install the library, run the following
#
# python setup.py install
#
# prerequisite: setuptools
# http://pypi.python.org/pypi/setuptools

PYTHON_REQUIRES = ">=3.9"
REQUIRES = [
    "lakefs-sdk >= 1.25, < 2",
    "pyyaml ~= 6.0.1",
]
TEST_REQUIRES = [
    "pytest ~= 7.4.3",
    "pytest-datafiles ~= 3.0.0",
    "pandas ~= 2.1.4",
    "pyarrow ~= 14.0.1",
    "pillow ~= 10.2.0"
]

with open('README.md') as f:
    long_description = f.read()

setup(
    name=NAME,
    version=VERSION,
    description="lakeFS Python SDK Wrapper",
    author="Treeverse",
    author_email="services@treeverse.io",
    url="https://github.com/treeverse/lakeFS/tree/master/clients/python-wrapper",
    keywords=["OpenAPI", "OpenAPI-Generator", "lakeFS API", "Python Wrapper"],
    python_requires=">=3.9",
    install_requires=REQUIRES,
    tests_require=TEST_REQUIRES,
    packages=find_packages(exclude=["tests"]),
    include_package_data=True,
    license="Apache 2.0",
    long_description=long_description,
    long_description_content_type='text/markdown',
    extras_require={
        "all": ["boto3 >= 1.26.0"],
        "aws-iam": ["boto3 >= 1.26.0"],
    },
)



from tests.utests.common import expect_exception_context
from lakefs.namedtuple import LenientNamedTuple


class NamedTupleTest(LenientNamedTuple):
    field1: str
    field2: int
    field3: bool


def test_namedtuple():
    nt = NamedTupleTest(
        field1="1",
        field2=1,
        field3=True)
    assert nt.field1 == "1"
    assert nt.field2 == 1
    assert nt.field3

    # try to modify value
    with expect_exception_context(AttributeError):
        nt.field1 = "2"

    # Create named tuple with unknown args
    kwargs = {
        "field1": "test",
        "field2": 2,
        "field3": False,
        "field4": "something",
        "field5": None,
    }
    nt2 = NamedTupleTest(**kwargs)
    assert nt.field1 != nt2.field1
    assert nt2.field1 == "test"
    assert nt2.field2 == 2
    assert not nt2.field3

    with expect_exception_context(AttributeError):
        nt2.field4  # pylint: disable=pointless-statement

    with expect_exception_context(AttributeError):
        nt2.field5  # pylint: disable=pointless-statement

    # Verify extra kwargs are in 'unknown' dict
    assert nt2.unknown['field4'] == "something"
    assert nt2.unknown['field5'] is None

    # Missing args
    with expect_exception_context(TypeError):
        NamedTupleTest(field2=1)

    # Verify comparison works with different unknown fields
    kwargs["field5"] = "something"
    nt3 = NamedTupleTest(**kwargs)
    assert nt2 == nt3



import os
import re
import time
import uuid
from pathlib import Path

import pytest

from lakefs import client
from lakefs.repository import Repository

TEST_STORAGE_NAMESPACE_BASE = os.getenv("STORAGE_NAMESPACE", "").rstrip("/")


def get_storage_namespace(test_name):
    return f"{TEST_STORAGE_NAMESPACE_BASE}/{uuid.uuid1()}/{test_name}"


def _setup_repo(namespace, name, default_branch):
    clt = client.Client()
    repo_name = name + str(int(time.time()))
    repo = Repository(repo_name, clt)
    repo.create(storage_namespace=namespace, default_branch=default_branch)
    return clt, repo


@pytest.fixture(name="test_name")
def fixture_test_name(request):
    return re.sub(r'[_\[\]]', "-", request.node.name.lower())


@pytest.fixture(name="storage_namespace")
def fixture_storage_namespace(test_name):
    return get_storage_namespace(test_name)


@pytest.fixture()
def setup_repo(storage_namespace, test_name, default_branch="main"):
    return _setup_repo(storage_namespace, test_name, default_branch)


@pytest.fixture(scope="session")
def setup_branch_with_commits():
    _, repo = _setup_repo(get_storage_namespace("branch-with-commits"),
                          "branch-with-commits",
                          "main")
    branch = repo.branch("test_branch").create("main")
    commit_num = 199
    for i in range(commit_num):
        obj = branch.object("test1")
        if not i % 2:
            obj.upload("test_data")
        else:
            obj.delete()
        branch.commit(f"commit {commit_num - i - 1}")
    return branch


@pytest.fixture(name="pre_sign", scope="function")
def fixture_pre_sign(request):
    clt = client.Client()
    if request.param and not clt.storage_config.pre_sign_support:
        pytest.skip("Storage adapter does not support pre-sign mode")
    return request.param


FIXTURE_DIR = Path(__file__).parent.parent.resolve() / 'test_files'
TEST_DATA = pytest.mark.datafiles(
    FIXTURE_DIR / 'data.csv',
    FIXTURE_DIR / 'data.json',
    FIXTURE_DIR / 'data.yaml',
    FIXTURE_DIR / 'data.xml',
    FIXTURE_DIR / 'data.parquet',
    FIXTURE_DIR / 'waving-axolotl.gif',
)
